#

'''
新的，更简单的buildz._dz.conf

简单的方便读取配置文件的类
没安全措施
每个Conf里有当前的数据conf，和数据是否存在exist
可以执行的操作:
读取，循环读取，修改，删除，修改和删除不会通知其他节点

push和pop操作(耗时高)
push修改当前conf的key
生成修改记录

pop的时候全改回来
因为没用到，暂时不实现，只保留接口和数据

gets可以获取多个key，返回队列
top(domain)回到根节点的domain下
call(domain)到当前节点的domain下，不存在或没有数据就报错

Conf包含数据：
conf: 实际数据
exist: 数据是否存在
domain: 根节点到当前节点的路径字符串
spt,spts: 分割字符串
src: 只读配置，默认为空
root: 根节点



'''

from . import mapz
from buildz import xf
from ..base import Base
import os
def dzkeys(key, spt):
    if key is None:
        return []
    elif type(key)==str:
        key = key.split(spt)
    elif type(key) not in (list, tuple):
        key = [key]
    return key

class Conf(Base):
    def len(self):
        return len(self.conf)
    def items(self, as_conf=True):
        rst = []
        for i in range(self.len()):
            rst.append(self.list(i, as_conf))
        return rst
    def list(self, i, as_conf=True):
        conf = self.conf[i]
        if as_conf:
            domain = self.domain+self.spt+str(i)
            val = conf
            src = self.src_list(i, as_conf)
            obj = self.root or self
            find = True
            conf = Conf(self.spt, self.spts, domain, obj, src, val, find)
        return conf
    def lhget(self, key, default=None, loop=-1):
        '''
            循环读取，直到读取到的不是字符串或loop==0或没有对应的key了
        '''
        a,b = self._hget(key, default)
        bak = a, b
        obj = self.root or self
        while b and type(a)==str and loop!=0:
            a,b = obj._hget(a, default)
            if b:
                bak = a,b
            loop-=1
        if loop>0:
            bak = a,b
        return bak
    def key(self, ks):
        return self.spt.join(ks)
    def src_list(self, i, as_conf = True):
        if self.src is None:
            return None
        if self.src.get_type()!=list:
            return None
        if self.src.len()<=i:
            return None
        return self.src.list(i, as_conf)
    def src_hget(self, key, default):
        if self.src is None:
            return default, 0
        return self.src.hget(key, default, loop)
    def src_dm(self, domain):
        if self.src is None:
            return None
        return self.src(domain,0)
    def src_top(self, domain):
        if self.src is None:
            return None
        return self.src.top(domain, 0)
    def lget(self, key, default=None, loop=-1):
        return self.lhget(key, default, loop)[0]
    def hget(self, key, default=None):
        '''
            最基本的读取方法
            先从conf拿，没有则从src拿
        '''
        val, find = default, 0
        keys = dzkeys(key, self.spt)
        if self.exist and type(self.conf)==dict:
            val, find = mapz.dget(self.conf, keys, default)
        if find:
            return val, find
        return self.src_hget(keys, default)
    def get(self, key, default=None):
        val = self.hget(key, default)[0]
        #print(f"get {key} = {val}")
        return val
    def has(self, key):
        return self.hget(key, None)[1]
    def remove(self, key):
        if not self.exist or type(self.conf)!=dict:
            return
        keys = dzkeys(key, self.spt)
        return mapz.dremove(self.conf, keys)
    def spts_ks(self, keys):
        '''
            一堆key组成的字符串拆分成key的列表
        '''
        keys = dzkeys(keys, self.spts)
        keys = [k.strip() if type(k) == str else k for k in keys]
        return keys
    def set(self, key, val):
        keys = dzkeys(key, self.spt)
        mapz.dset(self.conf, keys, val)
    def top(self, domain = None, loop=0):
        root = self.root or self
        if domain is not None:
            root = root(domain, loop)
        return root
    def l(self, domain=None, loop=-1):
        return self(domain,loop)
    def call(self, domain=None, loop=0):
        '''
            获取子域下的数据作为Conf
        '''
        if domain is None:
            return self.root or self
        val,find = self.hget(domain)
        src = self.src_dm(domain)
        if self.domain:
            domain = self.domain+self.spt+domain
        obj = self.root or self
        bak = domain, val, find, src
        while loop!=0 and find and type(val)==str:
            bak = domain, val, find, src
            val,find = obj.hget(domain)
            src = self.src_top(domain)
            loop-=1
        domain,val,find,src=bak
        return Conf(self.spt, self.spts, domain, obj, src, val, find)
    def val(self):
        return self.conf
    def get_type(self):
        return type(self.val())
    def has_val(self):
        return self.exist
    def get_conf(self):
        return self.conf
    def str(self):
        return str(self.get_conf())
    def init(self, spt='.', spts=',', domain=None, root = None, src = None, conf=None, exist=1):
        self.spt = spt
        self.spts = spts
        self.domain = domain
        self.root = root
        self.src = src
        if conf is None:
            conf = {}
        self.conf = conf
        self.exist = exist
        self.history = {}
    def clean(self):
        '''
            清空当前数据
        '''
        if self.get_type()!=dict:
            return
        self.history = {}
        for key in self.conf:
            del self.conf[key]
        return self
    def dkey(self, key):
        if self.domain:
            key = self.domain+self.spt+key
        return key
    def update(self, conf, flush = 1, replace=1, visit_list=0):
        if flush:
            conf = xf.flush_maps(conf, lambda x:x.split(self.spt) if type(x)==str else [x], visit_list)
        #print(f"update with {conf}")
        xf.fill(conf, self.conf, replace=replace)
        return self
    def push(self, key, value, flush = 1, update=0, clean_history = 0):
        if self.get_type()!=dict:
            return None
        keys = dzkeys(key, self.spt)
        val, find = mapz.dget(self.conf, keys)
        val = mapz.deep_clone(val)
        if clean_history or key not in self.history:
            self.history[key] = []
        self.history[key].append([val, find, update])
        if flush and type(value)==dict:
            value = xf.flush_maps(value, lambda x:x.split(self.spt) if type(x)==str else [x], 0)
        if find and type(value)==dict and type(val)==dict and update:
            self(key).update(value,flush=0)
        else:
            self.set(key, value)
        return key
    def pop(self, key, clean_history = 0):
        if key not in self.history:
            return False
        lst = self.history[key]
        if len(lst)==0:
            return False
        rst = lst.pop(-1)
        if clean_history:
            self.history[key] = []
        if not rst[1]:
            self.remove(key)
            return True
        self.set(key, rst[0])
        return True
    def pops(self, keys, *a, **b):
        keys = self.spts_ks(keys)
        keys.reverse()
        for key in keys:
            self.pop(key)
    def with_push(self, key, *a, **b):
        self.push(key, *a, **b)
        def out():
            self.pop(key)
        return pyz.with_out(out)
    def with_pushs(self, keys, *a, **b):
        self.pushs(keys, *a, **b)
        def out():
            self.pops(keys)
        return pyz.with_out(out)
    def g(self, **maps):
        rst = [self.get(k, v) for k,v in maps.items()]
        if len(rst)==1:
            rst = rst[0]
        return rst
    def s(self, **maps):
        [self.set(k,v) for k,v in maps.items()]
    def has_all(self, keys):
        keys = self.spts_ks(keys)
        rst = [1-self.has(key) for key in keys]
        return sum(rst)==0
    def has_any(self, keys):
        keys = self.spts_ks(keys)
        for key in keys:
            if self.has(key):
                return True
        return False
    @staticmethod
    def fcs_bind(fn, wfn, align=False, null_default= False):
        def wfc(self, keys, *objs, **maps):
            #print(f"[[[[[]]]]]{wfn} call by {id(self)}")
            keys = self.spts_ks(keys)
            fc = getattr(self, fn)
            rst = []
            for i in range(len(keys)):
                if i<len(objs):
                    val = fc(keys[i], objs[i], **maps)
                else:
                    if align:
                        raise Exception(f"not val[{i}]")
                    if null_default:
                        val = fc(keys[i], None, **maps)
                    else:
                        val = fc(keys[i], **maps)
                rst.append(val)
            return rst
        setattr(Conf, wfn, wfc)

pass
maps = xf.loads(r"""
(get,0,1)
(lget,0,1)
(set,1)
(remove)
(push,1)
""")
for item in maps:
    k = item[0]
    ks = k+"s"
    item = item[:1]+[ks]+item[1:]
    Conf.fcs_bind(*item)

pass




from .base import *
from ... import dz
from ..ioc.single import Single
class ObjectEncape(BaseEncape):
    key_single = Single.Key.default_param_key
    @staticmethod
    def set_single_id(params, id):
        params.set(ObjectEncape.key_single, id)
        return params
    '''
        id=?
        
    '''
    def init(self, single, src, args, maps, sets, before_set=None, after_set=None):
        super().init()
        self.single = Single(single,ObjectEncape.key_single)
        self.src, self.args, self.maps, self.sets = src, args, maps, sets
        self.before_set = before_set
        self.after_set = after_set
        self.objs = {}
    def call(self, params=None, **maps):
        obj, find = self.single.get(params)
        if find:
            return obj
        src = self.obj(self.src, params)
        if type(src)==str:
            src = self.load(src)
        args = self.elist2obj(self.args,params)
        _maps = self.edict2obj(self.maps,params)
        obj = src(*args, **_maps)
        obj_conf = Params.Clone(params, obj=obj)
        self.obj(self.before_set, obj_conf)
        sets = self.edict2obj(self.sets, params, False)
        for k,v in sets:
            setattr(obj, k, v)
        self.obj(self.after_set, obj_conf)
        self.single.set(params, obj)
        return obj
class ObjectDeal(BaseDeal):
    def deal(self, conf, unit):
        assert 'source' in conf
        id,id_find = unit.conf_key(conf)
        single = dz.g(conf, single=None)
        if single is None and not id_find:
            single = Single.Key.multi
        src = dz.g(conf, source=None)
        before_set, after_set = dz.g(conf, before_set=None, after_set=None)
        before_set = self.get_encape(before_set, unit)
        after_set = self.get_encape(after_set, unit)
        if Confs.is_conf(src):
            src = unit.get_encape(src, unit)
        elif type(src)==str:
            src = self.load(src)
        args = self.get_elist(unit, conf, 'args', [])
        maps = self.get_edict(unit, conf, 'maps', {})
        sets = self.get_edict(unit, conf, 'sets', [])
        encape = ObjectEncape(single, src, args, maps, sets,before_set,after_set)
        return encape

pass


#codig=utf-8
import os
import re
import base64
ACT_ADD = "add_prev"
ACT_DEL = "del_curr"
import struct
def encode(steps, as_str = False):
    """
        步骤列表转bytes/str
    """
    rst = b""
    # 加减的长度设置最多2**15-1，保留1位当作“加/减”，4位整数当字符串位置索引
    max16 = 2**15-1
    for step in steps:
        if step[0] == ACT_ADD:
            bs = step[2]
            if type(bs) == str:
                bs = bs.encode("utf-8")
            bs_arr = []
            while len(bs)>max16:
                cut = bs[:max16]
                bs = bs[max16:]
                bs_arr.append(cut)
            if len(bs)>0:
                bs_arr.append(bs)
            for bs in bs_arr:
                obj = struct.pack(">HI", len(bs)*2+1, step[1])+bs
                rst+=obj
        else:
            s1 = step[1]
            s2 = step[2]
            nexts = [s1]
            while (s2-s1)>max16:
                nexts.append(s1+max16)
                s1+=max16+1
            if s1<=s2:
                nexts.append(s2)
            for i in range(1,len(nexts)):
                obj = struct.pack(">HI", (nexts[i]-nexts[i-1])*2, nexts[i-1])
                rst+=obj
        #rst+=obj
    if as_str:
        rst = base64.b64encode(rst).decode()
    return rst

pass
def decode(obj):
    """
        bytes/str转步骤列表
    """
    if type(obj) == str:
        obj = base64.b64decode(obj)
    ni = 2+4
    rst = []
    while len(obj)>0:
        bs = obj[:ni]
        obj = obj[ni:]
        v2, base = struct.unpack(">HI", bs)
        act = v2%2
        c = v2//2
        #act, base, c = struct.unpack(">HI", bs)
        if act == 1:
            bs = obj[:c]
            obj = obj[c:]
            s = bs.decode("utf-8")
            tmp = [ACT_ADD, base, s]
        else:
            tmp = [ACT_DEL, base, base+c]
        rst.append(tmp)
    return rst

pass
def steps(vs, n, m, d, stra, strb):
    k = n-m
    rst = []
    for step in range(d, 0, -1):
        v = vs[step]
        x = v[k]
        y = x-k
        prev_v = vs[step-1]
        down = ((k == -step) or ((k != step) and prev_v[k + 1] > prev_v[k - 1]))
        prev_k = k+1 if down else k-1
        prev_x = prev_v[prev_k]
        prev_y = prev_x-prev_k
        if down:
            #加字符
            act = [ACT_ADD, prev_x, strb[prev_y]]
        else:
            #删字符
            act = [ACT_DEL, prev_x, prev_x, stra[prev_x]]
        rst.append(act)
        k = prev_k
    rst.reverse()
    rst = combine(rst)
    return rst

pass
def combine(steps):
    base = None
    rst = []
    for step in steps:
        if len(rst)>0 and rst[-1][0] == step[0]:
            prev = rst[-1]
            if step[0] == ACT_ADD:
                if prev[1]==step[1]:
                    prev[2]+=step[2]
                    continue
            elif step[0] == ACT_DEL:
                if prev[2] == step[1]-1:
                    prev[2] = step[2]
                    continue
            else:
                raise Exception("test")
        rst.append(step)
    return rst

pass
def count(steps):
    cnt = 0
    cadd,cdel = 0,0
    #print(f"steps: {len(steps)}")
    for step in steps:
        if step[0] == ACT_ADD:
            cnt+=len(step[2])
            cadd+=len(step[2])
        else:
            cnt+=1+step[2]-step[1]
            cdel +=1+step[2]-step[1]
    return cnt#,cadd,cdel

pass
def update(s, steps, split=1):
    """
        按照steps更新s
    """
    if split:
        if type(s)!=list:
            s = spt(s)
    if type(s)==bytes:
        s = [s[i:i+1] for i in range(len(s))]
    updated = -1
    rst = []
    spc = ""
    if type(s)==bytes or type(s[0])==bytes:
        spc = b""
    for step in steps:
        act = step[0]
        x = step[1]
        c = step[2]
        if updated+1<x:
            tmp = s[updated+1:x]
            rst.append(tmp)
        if act == ACT_ADD:
            rst.append(c)
            updated = x-1
        elif act == ACT_DEL:
            updated = c
        else:
            raise Exception("test")
    if updated+1 < len(s):
        rst.append(s[updated+1:])
    rst = [k if type(k)!=list else spc.join(k) for k in rst]
    rs = spc.join(rst)
    return rs

pass
def spt(s, spts = "\n'\" `~!@#$%^&*()_+-={}[]:;<>,.?/|"):
    arr = [s]
    spts = list(spts)
    if type(s)==bytes:
        spts = [k.encode() for k in spts]
    for _spt in spts:
        tmp = []
        for s in arr:
            s = s.split(_spt)
            # r = []
            # for k in s:
            #     r.append(k)
            #     r.append(_spt)
            # s = r[:-1]
            s = [k+_spt for k in s[:-1]]+[s[-1]]
            tmp+=s
        arr = tmp
    return arr

pass
def myers(stra, strb, split = 1):
    import time
    curr = time.time()
    if split:
        if type(stra)!=list:
            stra = spt(stra)
        if type(strb)!=list:
            strb = spt(strb)
    if type(stra)==bytes:
        stra = [stra[i:i+1] for i in range(len(stra))]
    if type(stra)==bytes:
        strb = [strb[i:i+1] for i in range(len(strb))]
    n = len(stra)
    m = len(strb)
    v = {
        1:0
    }
    vs = {
        0: {1:0}
    }
    index = 0
    total=0
    for d in range(n+m+1):
        tmp = {}
        for k in range(-d, d+1, 2):
            # k == -d, 左边界，没法从左边来，只能是上边往下，k==d,上边界，只能从左边过来，
            # v[k-1]是上一轮迭代，k-1线上的步数，从上一步的k-1到这一步的k，是向右，删字符，
            # 因为上一步算了k从-d+1到d-1，所以除了边界可能没有左右值，中间的k都能查到上一步的k+1和k-1
            # v[k+1]>v[k-1]，表示向下（加字符）的x比向右（删字符）的x大
            # 可以看图，k+1和k-1是k旁边的两条线，如果v[k+1]<=v[k-1]，说明k-1的点在k+1的下方或右边，看图会发现k-1这个点更接近k这条线的终点
            down = ((k == -d) or ((k != d) and v[k + 1] > v[k - 1]))
            kPrev = k+1 if down else k-1
            xStart = v[kPrev]
            yStart = xStart - kPrev
            # 完全没啥用的两个中间变量。。。
            xMid = xStart if down else xStart + 1
            yMid = xMid - k
            xEnd = xMid
            yEnd = yMid
            while xEnd < n and yEnd < m and stra[xEnd] == strb[yEnd]:
                xEnd+=1
                yEnd+=1
                total+=1
            v[k] = xEnd
            tmp[k] = xEnd
            index+=1
            total+=1
            if xEnd == n and yEnd == m:
              vs[d] = tmp
              sec = time.time()-curr
              stps = steps(vs, n,m,d,stra, strb)
              return stps
        vs[d] = tmp

pass

"""

python -m buildz myers


python -m buildz myers diff "D:\rootz\python\gits\buildz_upd\buildz\tz\myers_diff.py" "D:\rootz\python\wk\test\myers_diff.py" "D:\rootz\python\wk\test\stp.txt" +e


python -m buildz myers update "D:\rootz\python\gits\buildz_upd\buildz\tz\myers_diff.py" "D:\rootz\python\wk\test\myers_diff_.py" "D:\rootz\python\wk\test\stp.txt" +e

python -m buildz myers diff "D:\rootz\python\wk\test\myers_diff.py" "D:\rootz\python\wk\test\myers_diff_.py" "D:\rootz\python\wk\test\df.txt" +t

"""


#
from ..ioc.base import Base, EncapeData,IOCError
from ..ioc.single import Single
from ..ioc.decorator import decorator, IOCDObj
from .base import FormatData,FormatDeal
from buildz import xf, pyz
from buildz import Base as Basez
import os
#g_obj_cid = '_buildz_ioc_conf_index'
dp = os.path.dirname(__file__)
join = os.path.join
class ObjectDeal(FormatDeal):
    """
        对象object:
            {
                id: id
                type: object
                source: 导入路径+调用方法/类
                single: 1 //是否单例，默认是，
                        //这里的单例是一个id对应一个实例，
                        //如果两个id用的同一个source，就是同一个source的两个对象
                // 构造函数
                construct:{
                    args: [
                        item_conf,
                        ...
                    ]
                    maps: {
                        key1: item_conf,
                        ...
                    }
                }
                //construct和args+maps，不能同时存在
                args: [
                    item_conf,
                    ...
                ]
                maps: {
                    key1: item_conf,
                    ...
                }
                // sets之前调用方法
                prev_call: item_conf
                // 对象变量设置属性
                sets: [
                    {key: key1, data: item_conf }
                    ...
                ]
                // sets之后调用
                call: item_conf
                // remove前调用
                remove: item_conf
                // remove后调用
                after_remove: item_conf
            }
        简写：
            [[object, id, single], source, args, maps, sets, calls] 
                sets: [[key1, item_conf], ...]
        极简:
            [object, source]
        例:
            [object, buildz.ioc.ioc.conf.Conf] //生成Conf对象
    """
    def init(self, fp_lists = None, fp_defaults = None, fp_cst = None, fp_set = None):
        self.singles = {}
        self.single = Single("single", "id", 1)
        self.sources = {}
        super().init("ObjectDeal", fp_lists, fp_defaults, 
            join(dp, "conf", "obj_lists.js"),
            join(dp, "conf", "obj_defaults.js"))
        if fp_set is None:
            fp_set = join(dp, 'conf', 'obj_set_lists.js')
        if fp_cst is None:
            fp_cst = join(dp, 'conf', 'obj_cst_lists.js')
        self.fmt_set = FormatData(xf.loads(xf.fread(fp_set)))
        self.fmt_cst = FormatData(xf.loads(xf.fread(fp_cst)))
    def get_maps(self, maps, sid, id):
        if id is None:
            return None
        if sid not in maps:
            return None
        maps = maps[sid]
        if id not in maps:
            return None
        return maps[id]
    def set_maps(self, maps, sid, id, obj):
        if sid not in maps:
            maps[sid] = {}
        maps[sid][id] = obj
    def _deal(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        data = self.format(data)
        info = edata.info
        conf = edata.conf
        confs = edata.confs
        icst = None
        isets = None
        ivars = None
        if type(info) == dict:
            cid = xf.g(info, id=None)
            iargs, imaps = xf.g(info, args = None, maps = None)
            icst = {'args':iargs, 'maps':imaps}
            if iargs is None and imaps is None:
                icst = None
            isets = xf.g(info, sets = None)
            ivars = xf.g(info, vars=None)
        ids = self.single.get_ids(edata)
        id = xf.g(data, id = None)
        #print(f"obj.deal ids: {ids} for {data}")
        obj = self.single.get_by_ids(ids)
        if obj is not None:
            #raise IOCError(f"null for {ids}")
            return obj
        #source = xf.g(data, source=0)
        source = xf.g1(data, source=0, src=0)
        if source == 0:
            raise Exception(f"define object without 'source' key, {data}")
        source = self.get_obj(source, conf)
        if type(source)==str:
            fc = xf.get(self.sources, source, None)
        else:
            fc = source
        if fc is None:
            fc = pyz.load(source)
            self.sources[source]=fc
        cst = xf.g(data, construct = None)
        if cst is None:
            _args = xf.g(data, args = [])
            _maps = xf.g(data, maps = {})
            cst = [_args, _maps]
        cst = self.fmt_cst(cst)
        if icst is not None:
            xf.fill(icst, cst, 1)
        args = xf.g(cst, args=[])
        args = xf.im2l(args)
        maps = xf.g(cst, maps={})
        self.push_vars(conf, ivars)
        args = [self.get_obj(v, conf) for v in args]
        maps = {k:self.get_obj(maps[k], conf) for k in maps}
        obj = self.single.get_by_ids(ids)
        if obj is not None:
            return obj
        obj = fc(*args, **maps)
        self.single.set_by_ids(ids, obj)
        prev_call = xf.g(data, prev_call=None)
        if prev_call is not None:
            # TODO: 这边info透传不知道会不会有问题
            self.get_obj(prev_call, conf, obj, edata.info)
        sets = xf.g(data, sets=[])
        if type(sets)==list:
            tmp = {}
            for kv in sets:
                kv = self.fmt_set(kv)
                k = kv['key']
                v = xf.get_first(kv, "val", "data")
                tmp[k] = v
            sets = tmp
        if type(isets) == list:
            tmp = {}
            for kv in isets:
                kv = self.fmt_set(kv)
                k = kv['key']
                v = xf.get_first(kv, "val", "data")
                tmp[k] = v
            isets = tmp
        if isets is not None:
            xf.fill(isets, sets, 1)
        for k,v in sets.items():
            v = self.get_obj(v, conf, obj)#, edata.info)
            setattr(obj, k, v)
        call = xf.g(data, call=None)
        if call is not None:
            self.get_obj(call, conf, obj)#, edata.info)
        self.pop_vars(conf, ivars)
        #print(f"obj.deal ids: {ids} for {data}, rst: {obj}")
        return obj
    def remove(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        data = self.format(data)
        info = edata.info
        conf = edata.conf
        confs = edata.confs
        ids = self.single.get_ids(edata)
        obj = self.single.get_by_ids(ids)
        if obj is None:
            return None
        call = xf.g(data, remove=None)
        if call is not None:
            self.get_obj(call, conf, obj, edata.info)
        xf.removes(self.singles, ids)
        call = xf.g(data, after_remove=None)
        if call is not None:
            self.get_obj(call, conf, obj, edata.info)
        return None

pass
def update_set(maps):
    rst = {}
    for k,v in maps.items():
        if type(v)==str:
            v = xf.loads(v)
        rst[k] = v
    return rst

pass
def update_list(arr):
    rst = []
    for v in arr:
        if type(v)==str:
            v = xf.loads(v)
        rst.append(v)
    return rst

pass
ioc_conf_key = "_buildz_ioc_conf"
class IOCConf(Basez):
    def init(self, key = "_buildz_ioc_conf", ckey = "_buildz_ioc_conf_cls"):
        self.maps = {}
        self.cmaps = {}
    def get(self, cls, default):
        key = id(cls)
        if key not in self.cmaps:
            return default
        prv = self.cmaps[key]
        if prv!=cls:
            return default
        if key in self.maps:
            return self.maps[key]
        return default
    def unset(self, cls):
        key = id(cls)
        if key in self.cmaps:
            del self.cmaps[key]
            del self.maps[key]
    def set(self, cls, dt):
        key = id(cls)
        self.maps[key] = dt
        self.cmaps[key] = cls

pass
g_ioc_conf = IOCConf()
class IOCObjectAdd_(IOCDObj):
    def init(self, key, *arr):
        _arr = update_list(arr)
        self.key = key
        self._arr = _arr
    def call(self, cls):
        rst = {}
        rst = g_ioc_conf.get(cls, rst)
        _arr = []
        if self.key in rst:
            _arr = rst[self.key]
        _arr+=self._arr
        rst[self.key] = _arr
        g_ioc_conf.set(cls, rst)
        return cls

pass
class IOCObjectArgs(IOCObjectAdd_):
    def init(self, *arr):
        super().init("args", *arr)

pass
class IOCObjectMCall(IOCObjectAdd_):
    def init(self, *arr):
        # function, args, maps
        ks = "mcall,args,maps".split(",")
        _arr = []
        for item in arr:
            if type(item)==str:
                item = xf.loads(item)
            if type(item)==list:
                item = update_list(item)
                l = min(len(ks), len(item))
                _tmp = {}
                for i in range(l):
                    _tmp[ks[i]] = item[i]
                item = _tmp
            else:
                item = update_set(item)
            item['source'] = None
            item['type'] = "mcall"
            _arr.append(item)
        super().init("mcalls", *_arr)

pass
class IOCObjectSet_(IOCDObj):
    def init(self, key, **maps):
        _maps = update_set(maps)
        self._maps = {key:_maps}
    def call(self, cls):
        rst = {}
        rst = g_ioc_conf.get(cls, rst)
        xf.fill(self._maps, rst)
        g_ioc_conf.set(cls, rst)
        return cls

pass
class IOCObjectSet(IOCObjectSet_):
    def init(self, **maps):
        super().init("sets", **maps)

pass
class IOCObjectMap(IOCObjectSet_):
    def init(self, **maps):
        super().init("maps", **maps)

pass
class IOCObject(IOCDObj):
    KEYS = "id,args,maps,call,prev_call,single,remove,sets,after_remove,template,parent,temp".split(",")
    SET_KEYS = "maps,sets".split(",")
    def init(self, **maps):
        super().init()
        rst = update_set(maps)
        for key in self.SET_KEYS:
            if key not in maps:
                continue
            val = maps[key]
            for _k, _v in val.items():
                if type(_v)==str:
                    _v = xf.loads(_v)
                val[_k] = _v
        self._maps = rst
    def _set(self, key, **maps):
        sets = {}
        if key in self._maps:
            sets = self._maps[key]
        rst = update_set(maps)
        xf.fill(sets, rst)
        self._maps[key] = rst
        return self
    def map(self, **maps):
        return self._set("maps", **maps)
    def maps(self, **maps):
        return self.map(**maps)
    def set(self, **maps):
        return self._set("sets", **maps)
    def sets(self, **maps):
        return self.set(**maps)
    def call(self, cls):
        src = cls.__module__+"."+cls.__name__
        conf = {}
        conf = g_ioc_conf.get(cls, conf)
        xf.fill(self._maps, conf)
        conf['source'] = src
        conf['type'] = 'object'
        if 'mcalls' in conf and 'call' not in conf:
            conf['call'] = {'type': "calls", 'calls': conf['mcalls']}
        self.decorator.add_datas(conf)
        #conf[g_obj_cid] = self.decorator.add_datas(conf, self)
        g_ioc_conf.unset(cls)
        return cls

pass

#decorator.regist("IOCObject", IOCObject)
decorator.regist("obj", IOCObject)
decorator.regist("object", IOCObject)
decorator.regist("obj_set", IOCObjectSet)
decorator.regist("obj_map", IOCObjectMap)
decorator.regist("obj_arg", IOCObjectArgs)
decorator.regist("obj_mcall", IOCObjectMCall)
decorator.regist("obj_sets", IOCObjectSet)
decorator.regist("obj_maps", IOCObjectMap)
decorator.regist("obj_args", IOCObjectArgs)



'''
单元测试注解
'''
from ..base import Base,fcBase
from .. import logz
import traceback
'''
ut = UnitTest()
@ut.case("A")
def add(a,b):
    return a+b

@ut.case("A")
def sub(a,b):
    return a-b

@ut.case("B"):
def fv(v):
    return v*v

ut("A", [0,1])
ut("B", [2])

'''
class UnitTest(Base):
    def init(self, log=None):
        self.tests = []
        self.log = logz.make(log)
        self.bind_log('info')
        self.bind_log('debug')
        self.bind_log('warn')
        self.bind_log('error')
    def bind_log(self, key):
        def call(*a,**b):
            return getattr(self.log, key)(*a,**b)
        setattr(self, key, call)
    def case(self, key):
        key = str(key)
        def f(fc):
            self.tests.append([key, fc])
            return fc
        return f
    def wrap_call(self, _key, fc, args=[], maps={}):
        bk_log = self.log
        log = self.log(_key)
        self.log = log
        log.info("    unit test case: ", _key)
        rst = None
        done = True
        try:
            rst = fc(*args, **maps)
        except Exception as exp:
            done = False
            import traceback
            log.error("    unit test exp: ", exp)
            log.error(traceback.format_exc())
        log.info("    done unit test case: ", _key)
        self.log = bk_log
        return rst, done
    def single(self, key, args=[], maps={}):
        for _key, fc in self.tests:
            if _key==key:
                return self.wrap_call(_key, fc, args, maps)[0]
    def call(self, key = None, args=[], maps={}):
        if key is None:
            key = ''
        rst = []
        self.log.info(f"start unit test on key: {key}")
        for _key, fc in self.tests:
            if _key.find(key)==0:
                rst, done = self.wrap_call(_key, fc, args, maps)
                if not done:
                    break

pass
def make_loop_check(fc):
    def call(dts, args=[]):
        for i in range(len(dts)):
            dt = dts[i]
            fc(dt, args+[i])
    return call
def build(*a,**b):
    return UnitTest(*a,**b)


from buildz import argz,xf,Base,evalz,pathz
import sys,os

path = pathz.Path()
path.set("res", path(path.dir(__file__), "tests"))
conf = xf.loadf(path.res("conf.js"))
conf=xf.loads(r"""
type=ret
maps={
    spt: {
        src: [(env, spt), d]
        default: 1
        value: '.'
    }
    eglobal: {
        src: [(env, global), d]
        default:1
        value: true
    }
}
""")

mg = argz.CallBuilder()
fc = mg(conf)

data = xf.loads(r"""
args=[]
maps={
    env: {
        spt: x
    }
}
""")
args = xf.g(data, args=[])
maps = xf.g(data, maps={})

rst = fc(args, maps)
print(f"rst: {rst}")
'''
python -m buildz.argz.testx
python -m buildz.argz.test search /d/rootz f=asf c=test
'''



#

import re
from .jsc import fetch_methods_text, rps_dct, fetch_vals,fetch_vars
from .util import make_fcs

sc_dm = """
    <type> <method>(<params>) <exp>{
        <key> = dmKey(<key>);
        Conf obj = this;
        if (root != null) {
            obj = root;
        }
        <ret>obj.<call>(<vars>);
    }
""".rstrip()
# def domains(codes, args, text):
#     fcs, has_def = fetch_methods_text(codes, text)
#     rst = []
#     if has_def:
#         rst.append(codes.rstrip())
#     for _type, method, params in fcs:
#         rst+=domain(_type, method, params, args)
#     return rst
def domain(_type, method, params, exp, args):
    rst = []
    args = fetch_vals(args)
    ret = "return " if _type.find("void")<0 else ""
    if len(args)==0:
        wmethod = method[1:]
    else:
        wmethod = args[0]
    params = [k.strip() for k in fetch_vars(params)]
    vars = [k.split(" ")[-1].strip() for k in params]
    # key = 'key'
    # if len(args)>1:
    #     key =args[1]
    key = vars[0]
    rs = rps_dct(sc_dm, type =_type,method=wmethod,params=', '.join(params), key=key, call=method, vars = ', '.join(vars), ret = ret, exp=exp)
    rst.append(rs)
    #rst.reverse()
    return rst, []
domains = make_fcs(domain,False)



#
from ..tools import *
from buildz.ioc import wrap
#print(f"warp.obj_sets: {wrap.obj_sets}")
#raise Exception("test")
@wrap.obj(id="save")
@wrap.obj_args("ref, cache", "ref, log", "ref, cache.modify")
@wrap.obj_sets(cache_save = "ref, cache.save")
class Save(Base):
    def init(self, cache, log, upd):
        self.cache = cache
        self.log = log.tag("Save")
        self.upd = upd
    def call(self, data, next_fc=None):
        data = self.upd(data)
        save = xf.g(data, save={})
        fsave = xf.get(data, "save.file", {})
        mem = xf.get(data, "save.mem", {})
        saves = [save, fsave, mem]
        fcs = [self.cache.set, self.cache.set_file, self.cache.set_mem]
        for save,fc in zip(saves,fcs):
            for k, v in save.items():
                if type(v)==str:
                    v = ["key"]+v.split(".")
                if type(v)!=list:
                    v = ["key", v]
                tp = v[0]
                v = v[1:]
                if tp=="eval":
                    val = eval(v[0])
                elif tp == "exec":
                    exec(v[0])
                    val = self.val
                else:
                    val = xf.gets(data, v)
                fc(k, val)
        save = saves[0]
        if len(save)+len(fsave)>0:
            sv = self.cache_save.save()
            self.log.debug(f"cache_save in save: {sv}")
        return next_fc(data)

pass





from .dz import Conf
from . import xf, pathz, pyz, argx, dz
import sys, os
def load_conf(conf, dp=None, dp_key = 'dp', src_key = 'src.conf'):
    if type(conf)==dict:
        conf = Conf().update(conf)
    fps, base = conf.gets('fps, conf', [], {})
    if type(fps)==str:
        fps = [fps]
    conf_first,replace,flush,visit_list = conf.gets('conf_first,replace,flush,visit_list',1, 1,1,0)
    spt, spts = conf.gets('spt, spts','.',',')
    dp = conf.get(dp_key, dp)
    conf.set(dp_key, dp)
    path = pathz.Path()
    path.set("dp", dp)
    rst = Conf(spt, spts)
    if src_key is not None:
        rst.set(src_key, conf)
    if conf_first:
        rst.update(base, flush, replace, visit_list)
    for fp in fps:
        tmp = xf.loadf(path.dp(fp))
        if type(tmp)!=dict:
            continue
        rst.update(tmp, flush, replace, visit_list)
    if not conf_first:
        rst.update(base, flush, replace, visit_list)
    return rst
# using
def calls(conf):
    calls = conf.get("calls", [])
    local = conf.get("local", False)
    root = conf.top()
    if type(calls)==dict:
        target = dz.g(calls, target='run')
        calls = dz.get(calls, target, [])
        # if target in calls:
        #     calls = dz.get(calls, target, [])
        # else:
        #     calls = root.get(target, [])
    if type(calls)==dict:
        dm, init, calls, init_cover = dz.g(calls, domain=None, init = {},calls=[], init_cover=False)
        if len(init)>0:
            init_conf = conf.top("confz.init")
            if dm is not None:
                init_conf = init_conf(dm)
            init_conf.update(init, replace=init_cover)
            root.update(init_conf.val())
    if type(calls)==str:
        calls = [calls]
    obj = conf
    if not local:
        obj = root
    for key in calls:
        #assert obj.has(key), f"not has key: '{key}'"
        simple(obj.l(key))
    return conf
fn_key = "confz.fns"
fn_cache_key = "confz.fn.caches"
default_fn_key = "confz.fn.default"
def fn2fc(conf, key = 'fn', default_fn = None, default_fc = None):
    fn = conf.get(key)
    conf = conf()
    if fn is None:
        if default_fc is not None:
            return default_fc
        fn = default_fn or conf.get(default_fn_key, 'calls')
    fc = conf(fn_cache_key).get(fn)
    if fc is not None:
        return fc
    path =conf(fn_key).get(fn)
    if path is None:
        return None
    fc = pyz.load(path)
    conf(fn_cache_key).set(fn,fc)
    return fc
def get_fc(conf, fc_key='fc', fn_key='fn', default_fn = None, default_fc = None):
    fc = conf.get(fc_key)
    if fc is None:
        fc = fn2fc(conf, fn_key, default_fn, default_fc)
    else:
        fc = pyz.load(fc)
    return fc
def conf_update(conf):
    if conf.get_type()==str:
        conf = conf.ltop(conf.domain)
    up = conf.get('up', link=0)
    return conf
def deep_link(conf):
    up = conf.get('up', link=0)
    if not up:
        return
    deep_link(conf.top(up))
    conf().link(conf.domain, up)
    #conf.remove('up')
def deep_copy(conf):
    src = conf.get('copy', link=0)
    if not src:
        return
    deep_copy(conf.top(src))
    conf.update(conf.top(src).val(),replace=0)
    #conf.remove('copy')
def simple(conf):
    if conf.get_type()==str:
        conf = conf.ltop(conf.domain)
    deep_link(conf)
    deep_copy(conf)
    fc = get_fc(conf, default_fn = "calls")
    if fc == calls and not conf.has("calls"):
        fc = None
    if fc is None :
        fc = pyz.load(conf.domain)
    assert fc is not None, f"conf has not setted deal fc: {conf}"
    return fc(conf)
def get_sys_conf(conf = []):
    if type(conf) == str:
        conf = xf.loadf(conf)
    if conf is None:
        conf = []
    fetch = argx.Fetch(*conf)
    return fetch()
def fc_set(conf):
    """
        fn: set
        // domain: default=None
        // replace: default=1
        // flush: default=1
        conf: {

        }
    """
    maps,domain,flush,replace = conf.gets("conf, domain,flush,replace",None,None,1,1)
    if maps is None:
        return conf
    conf.top(domain).update(maps, flush=flush, replace=replace)
    return conf
def judge_fc(conf):
    fc = get_fc(conf, "judge_fc", "judge")
    rst = fc(conf)
    if rst:
        if not conf.has("yes"):
            return None
        conf = conf.l("yes")
    else:
        if not conf.has("no"):
            return None
        conf = conf.l("no")
    return simple(conf)
def switch_fc(conf):
    fc = get_fc(conf, "cal_fc", "cal", "get")
    val = fc(conf)
    deal_conf = conf("vals")(val)
    deal = get_fc(conf, "deal_fc", "deal", "mset")
    return deal(deal_conf)
def sub_conf(conf):
    deal = get_fc(conf, "deal_fc", "deal", "mset")
    val = conf("val")
    return deal(val)
def msets(conf, keys = []):
    # if not conf.has_val():
    #     return conf
    for k,v in conf.val().items():
        ks = keys+[k]
        if type(v)==str:
            val = conf.top(v).val()
            conf().set(conf.key(ks), val)
            #conf().top(conf.key(ks)).update(val)
            pass
        else:
            msets(conf(k), ks)
            #conf().top(k).update(v)
    return conf
def init_fn(conf):
    maps = {
        "calls": calls,
        "set": fc_set,
        "judge": judge_fc,
        "switch": switch_fc,
        "has_set": lambda conf:conf.top().has(conf.get("key")),
        "equal": lambda conf:conf.top().get(conf.get("key"))==conf.get("val"),
        "get": lambda conf: conf.top().get(conf.get("key")),
        "mset": lambda conf:conf.top().update(conf.val()) if conf.has_val() else conf,
        "deal_val":sub_conf,
        "msets": msets
    }
    conf(fn_cache_key).update(maps, replace=False)
def run(dp = None, fp = None, init_conf = {}):
    if dp is None:
        dp = os.path.dirname(__file__)
    path = pathz.Path()
    path.set('dp', dp)
    conf = {}
    if fp is not None:
        conf = xf.loadf(path.dp(fp))
    #sys_conf = get_sys_conf()
    xf.fill(init_conf, conf, 1)
    conf = Conf().update(conf)
    init = conf.get(conf.get("key.init", "init"), {})
    conf = load_conf(conf, dp).update(init)
    conf.set("confz.init", init)
    init_fn(conf)
    return simple(conf)
    #return conf
def test():
    run()
pyz.lc(locals(), test)



from . import proxy,mhttp,mhttps
import socket, threading
log = mhttp.log
from buildz import xf
import ssl
from os.path import join, isfile
import traceback
from buildz import fz

cert_conf = xf.loads(r"""
// 国家
country: CN
// 省份
provice: fujian
// 城镇
local: quanzhou
// 机构
org: testz
// 域名
comman: testz
// 邮箱
email: netz@testz
// 是否是ca
ca: true
//dns: [localhost]
// 签名有效提前时间（签名多少天前就有效）
valid_before: 1
// 签名有效时间（签名后有效多少天）
valid: 3650
dns: []
""")
sign_conf = xf.loads(r"""
// 签名有效提前时间（签名多少天前就有效）
valid_before: 1
// 签名有效时间（签名后有效多少天）
valid: 3650
""")
class CapsDealer(proxy.ProxyDealer):
    """
        http和https抓包，要自己实现MsgRecord类处理抓包数据，否则默认只是打印抓包的url和头部数据等信息
        buildz.netz.mhttp.record.MsgRecord的实现可参考buildz.netz.mhttp.record.MsgLog
    """
    def init(self, skt, ca, srv_context, channel_read_size=1024000, record=None,dp_cache=None):
        if dp_cache is None:
            dp_cache = "./res/certs"
        super().init(skt, channel_read_size, record=record)
        self.ca = ca
        self.srv_context = srv_context
        self.contexts = {}
        self.dp_cache = dp_cache
    def sign(self, hostname):
        dp = join(self.dp_cache, "_"+hostname)
        fz.makedir(dp)
        fp_csr = join(dp, "csr")
        fp_cert = join(dp, "cert")
        fp_prv = join(dp, "prv")
        if isfile(fp_cert):
            return fp_cert, fp_prv, None
        try:
            from buildz.netz import sslz
        except:
            # 不能创建域名证书，则直接用根证书
            return self.ca
        conf = dict(cert_conf)
        conf['comman']=hostname
        conf['dns'] = [hostname]
        sslz.gen_prv(fp_prv)
        sslz.gen_csr(fp_csr, fp_prv, conf, None)
        sslz.sign_csr(fp_cert, fp_csr, self.ca[0], self.ca[1],sign_conf, self.ca[2])
        return fp_cert, fp_prv, None
    def context(self, hostname):
        if hostname not in self.contexts:
            fp_cert, fp_prv, pwd = self.sign(hostname)
            self.contexts[hostname]=mhttps.load_server_context(fp_cert, fp_prv, pwd)
        return self.contexts[hostname]
    def deal_channel(self, skt_cli, skt_srv):
        #self.wskt.closefile()
        #wskt.closefile()
        hostname = skt_srv.addr[0]
        context = self.context(hostname)
        skt_cli = mhttp.WSocket.Bind(context.wrap_socket(skt_cli.skt, server_side=True))
        #context = ssl._create_unverified_context(ssl.PROTOCOL_TLS_CLIENT)
        skt_srv = mhttp.WSocket.Bind(self.srv_context.wrap_socket(skt_srv.skt, server_side=False))
        self.record.set_ssl(True)
        try:
            while True:
                if not skt_cli.readable():
                    continue
                line, headers, data_size = mhttp.http_recv(skt_cli)
                if line is None:
                    continue
                self.deals[None](skt_cli, line, headers, data_size, skt_srv)
        except Exception as exp:
            log.debug(f"channel exp: {exp}")
            log.warn(f"traceback: {traceback.format_exc()}")
        finally:
            self.record.set_ssl(False)


class CapsProxy(proxy.Proxy):
    """
        http和https抓包，要自己实现MsgRecord类处理抓包数据，否则默认只是打印抓包的url和头部数据等信息
        buildz.netz.mhttp.record.MsgRecord的实现可参考buildz.netz.mhttp.record.MsgLog
    """
    def init(self, addr, fp_sign, fp_prv, password = None,listen=5, record=None,cafile=None, capath=None, cadata=None,check_hostname=True, dp_cache = None):
        super().init(addr, listen, record)
        # context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        # context.load_cert_chain(fp_sign, fp_prv, password=password)
        self.ca = fp_sign, fp_prv, password
        self.srv_context = mhttps.load_verify_context(cafile, capath, cadata,check_hostname)
        if dp_cache is None:
            dp_cache = "./res/certs"
        self.dp_cache = dp_cache
    def make_dealer(self, skt, addr):
        return CapsDealer(skt, self.ca, self.srv_context, record=self.record.clone(), dp_cache = self.dp_cache)



from buildz.tls import *

class Server(Base):
    def init(self, conf, inst):
        addr = tuple(conf.gets("ip,port", "127.0.0.1", 8080))
        self.addr = addr
        self.make_deal = confz.get(inst,"server.deal")
        skt = socket.socket()
        skt.listen(addr)
        self.skt = skt
        self.running = True
        self.thread = None
    def stop(self):
        self.running = False
    def start(self):
        t = th.Thread(target=self.listen, daemon=True)
        t.start()
        self.thread = t
    def join(self):
        self.thread.join()
    def listen(self):
        self.running = True
        while self.running:
            skt = self.skt.accept()
            if not skt:
                continue
            t = th.Thread(target=self.make_deal(skt, self),daemon=True)
            t.start()
            self.tds.append(t)
        for t in self.tds:
            t.join()
    def call(self):
        self.listen()
pass





from ... import pyz, dz
from .tdata import TagData
class Datas(TagData):
    '''
        一个单元下的数据集
    '''
    def init(self, ns=None, id=None, dts=None):
        super().init(ns, id)
        self.dts = dts
    def bind(self, dts):
        if self.dts==dts:
            return
        self.dts = dts
        dts.add(self)
    def update(self, maps, tag=None):
        for key, val in maps.items():
            self.set(key, val, tag)
    def set(self, key, val, tag=None):
        tag = pyz.nnull(tag, self.default)
        tag = TagData.Key.stand(tag)
        super().set(key, val, tag)
        if self.dts is not None and tag in (TagData.Key.Pub, TagData.Key.Ns):
                self.dts.set(key, val, self.ns, tag, self)
    def tget(self, key, src=None,id=None, gfind=True):
        ns, id = self.nsid(src, id)
        obj, tag, find=super().tget(key, ns, id)
        if not find and gfind:
            obj, tag, find = self.dts.tget(key, ns, id)
        return obj, tag, find



#
import re,sys
from buildz import fz, dz, xf, argx, pyz, Base
def rps_dct(s, **maps):
    for k,v in maps.items():
        s = s.replace(f"<{k}>", v)
    return s

def rps_lst(s, *args):
    for i in range(0, len(args), 2):
        k,v = args[i:i+2]
        s = s.replace(f"<{k}>", v)
    return s
pt_param = ""
def has_body(s):
    return s.find("(")>=0
fc_pt = r'((?:public|protected|private){0,1}(?:static|\s)*(?:\<\w+\>){0,1}\s*[\w\<\>\[\]]*)\s+([\w]+)\s*\(([^)]*)\)\s*((?:throws\s+[^\{\}\[\]\(\)\;]+){0,1})\s*\{'
def fetch_method(s):
    pt = fc_pt#r'((?:public|protected|private|static|\s)*[\w\<\>\[\]]*)\s+([\w]+)\s*\(([^)]*)\)\s*\{'
    _type, method, params, exp = [k.strip() for k in re.findall(pt, s)[0]]
    return _type, method, params,exp
def fetch_methods(s):
    pt = fc_pt#r'((?:public|protected|private|static|\s)*[\w\<\>\[\]]*)\s+([\w]+)\s*\(([^)]*)\)\s*\{'
    return re.findall(pt, s)
def fetch_methods_text(s, text):
    #print(f"fetch_methods_text:", s)
    s=s.strip()
    mark_def = s.find("(")>0
    if not mark_def:
        fcs = fetch_methods(text)
        fcs = [k for k in fcs if k[1]==s]
        assert len(fcs)>0, f"method not found: {s}"
        tmp = fcs#[0]
    else:
        tmp = [fetch_method(s)]
    #print("tmp:", tmp)
    return tmp, mark_def
    rst = list(tmp)+[mark_def]
    return rst
def fetch_method_text(s, text):
    fcs, mark_def = fetch_methods_text(s, text)
    rst = list(fcs[0])+[mark_def]
    return rst
def fetch_vals(s):
    rst = []
    if s is None:
        return rst
    quote = None
    ignore = False
    tmp = ''
    for c in s:
        if ignore:
            ignore = False
        elif c == '\\':
            ignore = True
        elif c =='"':
            if quote is None:
                quote = c
            elif quote==c:
                quote = None
        elif c == ',' and quote is None:
            rst.append(tmp.strip())
            tmp = ''
            continue
        tmp+=c
    tmp = tmp.strip()
    if len(tmp)>0:
        rst.append(tmp)
    return rst

def fetch_vars(s):
    rst = []
    if s is None:
        return rst
    quote = None
    count = 0
    ignore = False
    tmp = ''
    for c in s:
        if c =='<':
            count+=1
        elif c=='>':
            count-=1
        elif c == ',' and count==0:
            rst.append(tmp.strip())
            tmp = ''
            continue
        tmp+=c
    tmp = tmp.strip()
    if len(tmp)>0:
        rst.append(tmp)
    return rst


def fc(s, val):
    return s
class Deals(Base):
    def init(self):
        self.fcs = {}
    def add(self, key, fc):
        self.fcs[key] = fc
    def fetch(self, order):
        left = order.find("(")
        right = order.rfind(")")
        od, val = order, None
        if left>0:
            assert right>0
            od = order[:left]
            val = order[left+1:right]
        return od.strip(), val
    def call(self, arr, orders, text):
        if type(arr) not in (list, tuple):
            arr = [arr]
        if len(orders)==0:
            return arr, []
        od = orders.pop(0)
        od, val = self.fetch(od)
        arr, others = self.call(arr, orders, text)
        #print(f"[TESTZ] arr:", arr)
        fc = self.fcs[od]
        rst = []
        for s in arr:
            a,b = fc(s, val, text)
            rst+=a
            others +=b
            #rst+=fc(s, val, text)
        return rst, others
class DealCodes(Deals):
    def init(self, left="<<", right=">>", spt="||",define="||"):
        super().init()
        self.left = left
        self.right = right
        self.spt = spt
        self.define = define
    def update(self, s):
        left, right, spt, define= self.left, self.right, self.spt, self.define
        pt = f"{left}[\s\S]*?{right}"
        finds = re.findall(pt, s)
        text = s
        for find in finds:
            src = find
            #print(f"[MATCH]:", find)
            #continue
            find = find[len(left):-len(right)]
            if find.find(define)==0:
                find = find[len(define):]
                i = 0
                while find[i:i+len(define)]!=define:
                    i+=1
                while find[i:i+len(define)]==define:
                    i+=1
                i-=1
                spt = find[:i]
                find = find[i+len(define):]
                #print(f"[XXX] find after define:",find)
            arr = find.split(spt)
            orders = arr[:-1]
            arr = arr[-1:]
            rst, others = self(arr, orders, s)
            rst+=others
            rs = "\n".join(rst)
            s = s.replace(src, rs)
            spt = self.spt
        return s



from buildz import Base, Args
# class Args(Base):
#     def size(self):
#         return len(self.args)+len(self.maps)
#     def str(self):
#         return f"<Args args={self.args}, maps={self.maps}>"
#     @property
#     def lists(self):
#         return self.args
#     @lists.setter
#     def lists(self, val):
#         self.args=val
#     @lists.deleter
#     def lists(self):
#         del self.args
#     @property
#     def dicts(self):
#         return self.maps
#     @dicts.setter
#     def dicts(self, val):
#         self.maps=val
#     @dicts.deleter
#     def dicts(self):
#         del self.maps
#     def init(self, args, maps):
#         self.args = args
#         self.maps = maps

# pass


#
from ... import pyz,dz,Base

class Key:
    default_param_key = 'single_id'
    i_multi = 0
    i_single = 1
    i_param = 2
    multi='multi'
    single='single'
    param='param'
    str2int = {
        multi:i_multi,
        single: i_single,
        param: i_param
    }
    @staticmethod
    def s2i(key=None):
        if key is None:
            key = Key.i_single
        if key in Key.str2int:
            key = Key.str2int[key]
        return key
class Single(Base):
    Key = Key
    def init(self, single=None, param_key = None):
        if param_key is None:
            param_key = Key.default_param_key
        self.single = Key.s2i(single)
        self.objs = {}
        self.param_key = param_key
    def get_key(self, params):
        if self.single == Key.i_multi:
            return 0,0
        elif self.single == Key.i_single:
            key = None
        elif self.single == Key.i_param:
            if params is None:
                key = None
            else:
                key = params.get(self.param_key)
        else:
            assert 0
        return key,1
    def set(self, params, obj):
        key, find = self.get_key(params)
        if not find:
            return
        self.objs[key] = obj
    def get(self, params):
        key, find = self.get_key(params)
        if not find:
            return 0,0
        if key not in self.objs:
            return 0,0
        return self.objs[key],1


#coding=utf-8

from ..ioc.base import Base, EncapeData, IdNotFoundError
from buildz import xf
from ..ioc.decorator import decorator, IOCDObj
import os
dp = os.path.dirname(__file__)
join = os.path.join
default_deals = join(dp, 'conf', 'deals.js')
class FormatData(Base):
    """
    {
        id: ...
        type: object
        single: 1
        source: 
        construct: {
            args: []
            maps: {
            }
        }
        sets: [
            {key: ..., val: ..., type: ...}
        ]
    }
    [[id, type, single], source, construct:[args, maps], sets]
    [
        {
            nullable: 0
            out: 1
            //key: construct
            //default: []
            conf: [
                {key: id, default: null}
                {key: type, default: null }
                {key: single, default: 1}
            ]
        }
        {
            nullable: 0
            key: source
        }
        {
            key: construct
            default: {args: [], maps: {}}
            conf: [
                {key: args, default: []}
                {key: maps, default: {}}
            ]
        }
        {
            key: sets
            default: []
        }
    ]
    """
    def init(self, lists = [], defaults = {}, name = "demo"):
        self.lists = lists
        self.defaults = {}
        self.name = name
    def deal(self, data, lists = None, defaults = None):
        data = self.l2m(data, lists)
        data = self.fill(data, defaults)
        return data
    def fill(self, data, defaults = None):
        if defaults is None:
            defaults = self.defaults
        "没有值的时候用默认值填充"
        xf.deep_update(data, self.defaults, replace=0)
        return data
    def l2m(self, data, lists=None):
        if type(data) == dict:
            return data
        if type(data) not in [list, tuple]:
            data = [data]
        if lists is None:
            lists = self.lists
        if type(lists) in [list, tuple]:
            lists = {'sort':1, 'data': lists}
        maps = {}
        i=0
        sort = xf.g(lists, sort=1)
        lists = xf.g(lists, data=[])
        if sort:
            rs = range(len(lists))
        else:
            rs = range(len(lists)-1, -1, -1)
        #for obj in lists:
        cnt = 0
        for i in rs:
            obj = lists[i]
            key = xf.g(obj, key=None)
            out = xf.g(obj, out = 0)
            if not out and key is None:
                raise Exception(f"error format in {self.name}, has not key and not out")
            if cnt>=len(data):
                null = xf.g(obj, nullable=1)
                if not null:
                    raise Exception(f"error format in {self.name}, not default value for index {i}, key {key}")
                if "default" in obj:
                    default = xf.g(obj, default=None)
                    maps[key] = default
                cnt +=1
                continue
            next_conf = xf.g(obj, conf = None)
            #print(f"[TEST CALL] i: {i}, data: {data}, obj: {obj}")
            obj = data[i]
            if next_conf is None:
                maps[key] = obj
                cnt +=1
                continue
            _maps = self.l2m(obj, next_conf)
            if out:
                maps.update(_maps)
            else:
                maps[key] = _maps
            cnt +=1
        return maps

pass

class BaseDeal(Base):
    def push_vars(self, conf, vars):
        if vars is not None:
            [conf.push_var(key,val) for key,val in vars.items()]
    def pop_vars(self, conf, vars):
        if vars is not None:
            [conf.pop_var(key) for key in vars]
    """
        基础处理类，加了一些方便处理的方法，自己写的处理可以不用继承这个
        自己实现的处理类，要实现两个方法：__call__(self, edata:EncapeData)和remove(self, edata:EncapeData)
        其中remove可以只写个空方法
    """
    def get_data(self, data, conf, src = None, info = None):
        if type(data) not in [list, dict, tuple]:
            i = conf.confs.data_index_type[0]
            data = [conf.default_type(), data]
            if i != 0:
                data.reverse()
        _type = conf.confs.get_data_type(data, 1, conf.default_type())
        edata = EncapeData(data, conf, local=True, type=_type, src = src, info = info)
        return edata
    def fill_objs(self, data, conf, src = None, info = None):
        if type(data)==list:
            data = [get_obj(k, conf, src, info) for k in data]
        else:
            data = {k:get_obj(v, conf, src, info) for k,v in data.items()}
        return data
    def get_obj(self, data, conf, src = None, info = None):
        edata = self.get_data(data, conf, src, info)
        return edata()
        if type(data) not in [list, dict, tuple]:
            i = conf.confs.data_index_type[0]
            data = [conf.default_type(), data]
            if i != 0:
                data.reverse()
        _type = conf.confs.get_data_type(data, 1, conf.default_type())
        edata = EncapeData(data, conf, local=True, type=_type, src = src, info = info)
        return edata()
        # deal = conf.get_deal(edata.type)
        # if deal is None:
        #     return None
        # return deal(edata)
    def deal(self, edata:EncapeData):
        """
        """
        try:
            return self._deal(edata)
        except Exception as exp:
            if not isinstance(exp, IdNotFoundError):
                print(f"except in {self.__class__}.deal({edata.data}): {exp}")
            raise
        return None
    def _deal(self, edata:EncapeData):
        """
        """
        return None
    def remove(self, edata:EncapeData):
        return None

pass
class FormatDeal(BaseDeal):
    """
        格式化处理类，加了自动装填，但感觉不太好用（一方面忘记写注释，写的时间久了，不知道要怎么用了，另一方面不好扩展，再一方面自动填充要消耗时间））
    """
    def init(self, name = "FormatDeal", fp_lists = None, fp_defaults = None, df_fp_lists=None, df_fp_defaults=None):
        if fp_lists is None:
            fp_lists = df_fp_lists
        if fp_defaults is None:
            fp_defaults = df_fp_defaults
        lists = []
        defaults = {}
        if fp_lists  is not None:
            lists = xf.loads(xf.fread(fp_lists))
        if fp_defaults  is not None:
            defaults = xf.loads(xf.fread(fp_defaults))
        self.format = FormatData(lists, defaults, name)
    def fill(self, data):
        data = self.format.l2m(data)
        self.format.fill(data)
        return data
    # def get_obj(self, data, conf, src = None, info = None):
    #     if type(data) not in [list, dict, tuple]:
    #         i = conf.confs.data_index_type[0]
    #         data = [conf.default_type(), data]
    #         if i != 0:
    #             data.reverse()
    #     _type = conf.confs.get_data_type(data, 1, conf.default_type())
    #     edata = EncapeData(data, conf, local=True, type=_type, src = src, info = info)
    #     return edata()
    #     deal = conf.get_deal(edata.type)
    #     if deal is None:
    #         return None
    #     return deal(edata)

pass

class Order(IOCDObj):
    """
        应该是什么没写完全的功能，已经忘记是做啥用的了
    """
    def init(self, *fcs, reverse=False):
        self.fcs = fcs
        if reverse:
            self.fcs.reverse()
        self.reverse = reverse
    def deal(self, cls):
        for fc in self.fcs:
            cls = fc(cls)
        return cls

pass
decorator.regist("orders", Order)


#

from buildz import xf

s = r"""
a=0
b=1
c={
    a=2,
    d=3
    e=4
    f=5
    g=6
}
h=[
    {a=0,b=1,c=2}
    asdf
    'haha'
    [0,0,0]
]
"""
obj = xf.loads(s)
print("obj")
print(obj)

rst = xf.dumps(obj)
print("rst:")
print(rst)
print("\n")
rst = xf.dumps(obj,format=1)
print("rst:")
print(rst)
print("\n")
rst = xf.dumps(obj,format=1, not_format_height=2)
print("rst:")
print(rst)
print("\n")
rst = xf.dumps(obj,format=1, not_format_height=2,json_format=1)
print("rst:")
print(rst)
print("\n")

def build(n=30, m=6, l=6, val = 123):
    _arr = [val]
    print("test A")
    for i in range(n):
        _arr = [list(_arr)]
    print("test B")
    _map = {}
    for i in range(m):
        _map[i] = dict(_map)
    print("test C")
    rst = []
    for i in range(l):
        rst.append([_arr,_map])
    return rst

pass
from buildz.xf import write, writez
from buildz import xf
from buildz.tz import time as timez
timecost = timez.timecost(out_sec = 1)
dumpz = timecost(writez.dumps)
dump = timecost(write.dumps)
val = xf.loadf(r"D:\rootz\python\lx\pyz\docs\工作\zrecords\process.js")
obj = build(100,12,12, val)
rst, secz = dumpz(obj,format=1)
rst, sec = dump(obj,format=1)
print(f"writez = {secz/sec} write")
print(f"write = {sec/secz} writez")
#print("obj:", rst)


import sys
sys.path.insert(0,"/home/zzz/works/codes/python")
from buildz.netz.sslz import cert

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from cryptography.x509.oid import NameOID

root_prv = "root.prv"
test_prv = "test.prv"
cert.genf_prv(root_prv)
cert.genf_prv(test_prv)

cert.genf_root_cert("root.cert", cert.loadf_prv(root_prv))
cert.genf_csr("test.csr", cert.loadf_prv(test_prv))
cert.signf_csr("test.cert", cert.loadf_prv(root_prv), cert.loadf_csr("test.csr"), cert.loadf_cert("root.cert"))
vrf = cert.verifyf_certs("test.cert")
print(f"verify result: {vrf}")

#print(dir(csr))
#exit(0)
#csr.verify()




from .base import *
from ..ioc_deal import deal
class DealDeal(deal.DealDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.ikey(1, 'source', 'src'.split(","), need=1)
        conf.index(2, 'call')
        conf.ikey(3, 'deals', 'deal'.split(","))
        conf.index(4, 'tag')
        self.update = conf

pass


#
import re
from .jsc import fetch_method_text, rps_dct, fetch_vals,fetch_vars
from .util import make_fcs
sc_def = r'''
    <type> <method>(<params>) <exp>{
        <ret><call>(<args>);
    }
'''.rstrip()
#def default(codes, args, text):
def default(_type, method, params, exp, args):
    #_type, method, params, has_def = fetch_method_text(codes, text)
    rst = []
    if args is None or args.strip()=='':
        return rst, []
    arr = fetch_vals(args)
    call = method
    ret = 'return '
    if _type.strip() in ['','public','protected','private']:
        call = "this"
        ret = ''
    if _type.find("void")>=0:
        ret = ""
    params = [k.strip() for k in fetch_vars(params)]
    vars = [k.split(" ")[-1] for k in params]
    for i in range(1, len(arr)+1):
        defs = arr[-i:]
        _params = params[:-i]
        _vs = vars[:-i]
        _vs = _vs+defs
        _vs = ", ".join(_vs)
        _params = ", ".join(_params)
        s = rps_dct(sc_def, type=_type, method = method, params=_params, ret = ret, call=call, args = _vs, exp=exp)
        rst.append(s)
    return rst, []
defaults = make_fcs(default,True)


from .. import base
from .. import item
from .. import exp
from ... import file
from ... import code as codez
from . import lr

class PrevStrDeal(lr.LRDeal):
    def types(self):
        if not self.deal_build:
            return []
        return ['str']
    def build(self, obj):
        obj.is_val = 1
        if self.translate:
            val = obj.val
            val = self.do_translate(val)
            obj.val = val
        return obj
    def prepare(self, mg):
        super().prepare(mg)
        self.as_bytes = mg.as_bytes
        self.label_l2 = mg.like("\\")
        self.label_qt = mg.like('"')
        self.label_et = mg.like("\n")
        self.label_lr = mg.like("\r")
        self.label_nl = mg.like("")
        self.et_in_right = self.right.count(self.label_et)
    def init(self, left = '"', right= '"', single_line = False, note = False, translate = False, deal_build = False):
        super().init(left, right, 'str')
        self.single_line = single_line
        self.note = note
        self.translate = translate
        self.deal_build = deal_build
        self.as_bytes = True
        self.octs = None
        self.hexs = None
    def do_translate(self, s):
        """
            取巧用python的eval来生成字符表
        """
        is_bytes = type(s)==bytes
        if is_bytes:
            return codez.ubytes(s, "utf-8")
        else:
            return codez.ub2s(s.encode("utf-8"), "utf-8")
            return codez.ustr(s)
    def deal(self, buffer, rst, mg):
        cl = buffer.read(self.ll)
        if cl != self.left:
            return False
        rm = buffer.full().strip()
        rm_pos = buffer.pos()
        buffer.clean2read(self.ll)
        if len(rm)>0:
            if not self.note:
                raise exp.Exp(f"unexcept char before string: {rm}", rm_pos)
            else:
                rst.append(item.Item(rm, rm_pos, type = "str", is_val = 0))
        tmp = cl[:0]
        ctmp = tmp[:0]
        do_judge = 1
        mark_et = 0
        mark_l2 = 0
        while True:
            if do_judge and self.right == buffer.rget(self.lr):
                break
            c = buffer.read_cache(1)
            if do_judge and c == self.label_et:
                mark_et += 1
            if len(c)==0:
                if self.single_line and self.note:
                    break
                raise exp.Exp(f"unexcept string end while reading str", buffer.pos())
            do_judge = 1
            if c == self.label_l2:
                mark_l2 = 1
                do_judge = 0
                c = buffer.read_cache(1)
                if len(c)==0:
                    raise exp.Exp(f"unexcept string end while reading str", buffer.pos())
        data = buffer.full()
        data_pos = list(buffer.pos())
        data = data[:-self.lr]
        data_pos[1]-= self.lr
        data_pos = tuple(data_pos)
        buffer.clean()
        mark_et -= self.et_in_right
        if self.single_line and mark_et>0:
            print("left:",self.left, "right:", self.right)
            raise exp.Exp(f"contain enter in single line string",data_pos)
        if self.translate and mark_l2:
            data = self.do_translate(data)
        if self.note:
            return True
        rst.append(item.Item(data, data_pos, type='str', is_val = 1))
        return True

pass




from .. import xf
from .. import ioc
from ..base import Base
from ..ioc import wrap
class DealType(Base):
    def init(self, deals = {}, cache = None):
        self.deals = deals
        self.cache = cache
    def call(self, data):
        if type(data)==str:
            fp = data
            if self.cache is not None:
                fp = self.cache.rfp(fp)
            data = xf.loadf(fp)
        _type = xf.g(data, type = None)
        return self.deals[_type](data)

pass

@wrap.obj(id = "def.deal.type")
@wrap.obj_args("ioc, confs")
@wrap.obj_set(cache="ref, cache")
class DefDeal(Base):
    def init(self, mg):
        self.mg = mg
    def call(self, maps, fp):
        conf = xf.get(maps, "def.deal", None)
        factory_id = xf.g(conf, factory="buildz.auto.deal.fill")
        factory = self.mg.get(factory_id)
        data = xf.g(conf, types={})
        rst = {}
        for _type, calls in data.items():
            rst[_type] = factory(calls, True)
        obj = DealType(rst, self.mg.get("cache"))
        maps['deal_obj'] = obj
        return True

pass



#

import sys
from buildz.gpuz.torch import DictCache
from buildz.gpuz.test import analyze
from buildz import pyz
import torch,time
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
cpu,cuda = analyze.dvs
class ConvModel(nn.Module):
    def __init__(self, dims, num, ins_channels, middle_channels):
        super().__init__()
        nets=[]
        for i in range(num):
            nets.append(nn.Conv2d(ins_channels, middle_channels, 5, padding=2))
            nets.append(nn.Conv2d(middle_channels, ins_channels, 5, padding=2))
            nets.append(nn.LeakyReLU())
        self.nets = nn.Sequential(*nets)
    def forward(self, inputs):
        return self.nets(inputs)
    def size(self):
        return sum([analyze.unit_sz(net) for net in self.nets])

pass
class ResNet(nn.Module):
    def __init__(self, net):
        super().__init__()
        self.net = net
    def forward(self, inputs):
        return inputs+self.net(inputs)
class TestDataset(Dataset):
    def __init__(self, n, dims, channels):
        self.n = n
        self.dims = dims
        self.datas = torch.rand(n, channels, dims,dims)
        sz = analyze.sz(self.datas)
        sz, unit = analyze.show_size(sz)
        print(f"data size: {sz} {unit}")
    def __len__(self):
        return self.n
    def __getitem__(self, i):
        return self.datas[i], self.datas[i]

pass
def gen(dims, nets_num, channels, middle_channels, num_conv, lr):
    """
        类似UNet的模型
    """
    mds = []
    base = None
    for i in range(num_conv):
        nets = [ConvModel(dims, nets_num, channels, middle_channels) for i in range(2)]
        mds+=nets
        if base is not None:
            nets = [nets[0], base, nets[1]]
        nets = nn.Sequential(*nets)
        base = ResNet(nets)
    fullnet = base
    opts =[optim.Adam(md.parameters(), lr=lr) for md in mds]
    gopt = optim.Adam(fullnet.parameters(), lr=lr)
    mds_sz = [md.size() for md in mds]
    sz, unit = analyze.show_size(sum(mds_sz))
    print(f"Model Size: {sz} {unit}")
    return mds, fullnet, opts, gopt
    


def fc_opt(net, opt):
    torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
    opt.step()
def test():
    nets=2
    channels=10
    middle_channels = 30
    dims=512
    loop = 10
    datas = 6
    batch=2
    '''
        如果batch改成6，不用DictCache的时候8GB显存的显卡直接报错显存不够；用DictCache并且win_size=3时需要5GB显存，训练一次9秒；用cpu则要跑很久很久很久
    '''
    # batch=6
    lr=0.0001
    win_size=3
    num_conv = 12
    args = sys.argv[1:]
    mark_train = True
    if len(args)>0:
        mark_train = args.pop(0).lower()=='train'
    modes = 'cuda,cache,cpu'
    if len(args)>0:
        modes = args.pop(0)
    if len(args)>0:
        num_conv = int(args.pop(0))
    print(f"num_conv: {num_conv}")
    ds = TestDataset(datas, dims, channels)
    dl = DataLoader(ds, batch)
    loss_fn = torch.nn.MSELoss()
    def fc_gen():
        return gen(dims, nets, channels, middle_channels, num_conv, lr)
    analyze.analyzes(mark_train, loop, fc_gen, dl, loss_fn, fc_opt, win_size, modes)

pyz.lc(locals(),test)


#
from .base import *
from .datas import Datas
from .tdata import TagData, UnitBase
from .confs import Confs
from .encapes import Encapes
from ... import pyz
from .builds import Builds
from .envs import Envs
class Unit(UnitBase):
    def nsid(self, src=None, id=None):
        if src == -1:
            src = self
        if isinstance(src, UnitBase) or isinstance(src, TagData):
            src,id = src.ns, src.id
        return src, id
    def init(self, ns=None, deal_ns=None, env_ns=None, deal_key = None, conf_key = None, tag_key = None, id=None):
        super().init(ns, id)
        self.deal_ns = deal_ns
        self.env_ns = env_ns
        self.deal_key = deal_key
        self.conf_key = conf_key
        self.tag_key = tag_key
        self.confs = Confs(ns, deal_ns, id)
        self.deals = Datas(deal_ns, id)
        self.envs = Envs(env_ns, id)
        self.builds = Builds(self)
        self.mg = None
        self.build_encapes()
        self.vars = []
    def deal_vars(self):
        if self.mg is None or len(self.vars)==0:
            return
        for kvs, src, tag in self.vars:
            self.push_vars(kvs, src, tag)
        self.vars = []
    def add_build(self, conf):
        self.builds.add(conf)
    def build(self):
        self.builds.build()
    def is_conf(self, obj):
        return self.confs.is_conf(obj)
    def build_encapes(self):
        if self.confs is None or self.deals is None or self.deal_key is None:
            self.encapes = None
            return
        self.encapes = Encapes(self.ns, self.id, None, self)
    def bind(self, mg):
        if self.mg == mg:
            return
        self.mg = mg
        self.deal_key = pyz.nnull(self.deal_key, mg.deal_key)
        self.conf_key = pyz.nnull(self.conf_key, mg.conf_key)
        self.tag_key = pyz.nnull(self.tag_key, mg.tag_key)
        if self.id is None:
            self.id = mg.id()
            self.confs.set_id(self.id)
            self.deals.set_id(self.id)
            self.envs.set_id(self.id)
        self.mg.add(self)
        self.builds.bind(mg.builds)
        self.confs.bind(mg.confs)
        self.deals.bind(mg.deals)
        self.envs.bind(mg.envs)
        self.build_encapes()
        self.encapes.bind(mg.encapes)
        self.deal_vars()
    def lc_get_env(self, key, ns=None, tag=None, id=None):
        return self.envs.tget(key, ns, id, False)
    def get_env(self, key, src=-1, id=None, gfind=True):
        src, id = self.nsid(src, id)
        if not gfind:
            return self.lc_get_env(key, src, None, id)
        return self.mg.get_env(key, src, None, id, self.lc_get_env)
    def update_env(self, maps, tag=None, flush=False):
        self.envs.update(maps, tag, flush)
    def set_env(self, key, val, tag=None):
        self.envs.set(key, val, tag)
    def get_deal(self, key, src=-1, id=None, gfind=True):
        src, id = self.nsid(src, id)
        return self.deals.tget(key, src, id, gfind)
    def set_deal(self, key, deal, tag=None):
        self.deals.set(key, deal, tag)
    def get_conf(self, key, src=-1, id=None, gfind=True):
        src, id = self.nsid(src, id)
        return self.confs.tget(key, src, id, gfind)
    def set_conf(self, key, conf, tag=None):
        self.conf_key.fill(conf, key)
        self.confs.set(key, conf, tag)
    def tag(self, conf, _tag=None):
        if _tag is None:
            _tag, find = self.tag_key(conf)
        return _tag
    def add_conf(self, conf, tag = None):
        key,find = self.conf_key(conf)
        tag = self.tag(conf, tag)
        if find:
            self.set_conf(key, conf, tag)
    def get_encape(self, key, src=-1, id=None, gfind=True):
        src, id = self.nsid(src, id)
        self.mg.build()
        return self.encapes.tget(key, src, id, gfind)
    def set_encape(self, key, encape, tag=None):
        self.encapes.set(key, encape, tag)
    def get_var(self, key, src=-1, tag=None):
        src, id = self.nsid(src, None)
        return self.mg.get_var(key, src, tag)
    def push_vars(self, kvs, src=-1, tag=None):
        if self.mg is None:
            self.vars.append([kvs, src, tag])
            return
        src, id = self.nsid(src, None)
        return self.mg.push_vars(kvs, src, tag)
    def get(self, key, params=None, src=-1, id=None, gfind=True, search_var=True):
        src, id = self.nsid(src, id)
        if search_var:
            obj, find = self.mg.get_var(key, src, TagData.Key.Ns)
            if find:
                return obj, 1
        encape, _tag, find = self.get_encape(key, src, id, gfind)
        if not find:
            return None, 0
        return encape(params),1
    def get_obj(self, key, params=None, src=-1, id=None, gfind = True, search_var=True):
        return self.get(key, params, src, id, gfind, search_var)[0]


#coding=utf-8
import os
coding="utf-8"
def bread(filepath):
    if type(filepath) not in [str, bytes, os.PathLike]:
        raise Exception(f"error filepath: {filepath}")
    with open(filepath, 'rb') as file:
        s = file.read()
    return s

pass
def decode(s, coding = 'utf-8'):
    coding = coding.lower()
    xcoding = 'utf-8'
    if coding == 'utf-8':
        xcoding = 'gbk'
    try:
        return s.decode(coding)
    except:
        return s.decode(xcoding)

pass
def decode_c(s, coding = 'utf-8'):
    coding = coding.lower()
    xcoding = 'utf-8'
    if coding == 'utf-8':
        xcoding = 'gbk'
    try:
        return s.decode(coding), coding
    except:
        return s.decode(xcoding), xcoding

pass
def fread(filepath, coding = 'utf-8'):
    s = bread(filepath)
    return decode(s, coding)

pass
def fread_c(filepath, coding = 'utf-8'):
    s = bread(filepath)
    return decode_c(s, coding)

pass

def fwrite(content, filepath, mode = 'w'):
    with open(filepath, mode) as f:
        f.write(content)

pass



from . import Base
import os,sys,json,time
dirname = os.path.dirname
join = os.path.join
class Key:
    first='first'
    ifirst = 0
    last= 'last'
    ilast = -1
    @staticmethod
    def check_first(val):
        return val in (Key.first, Key.ifirst)
    @staticmethod
    def check_last(val):
        return val in (Key.last, Key.ilast)
class Path(Base):
    Key = Key
    def init(self, *args, **maps):
        self.paths = {}
        self.fcs = {}
        self.lasts = {}
        k = "check_abs"
        self.check_abs = True
        if len(args)>0:
            self.set(args)
        if k in maps:
            self.check_abs = maps[k]
            del maps[k]
        for k, v in maps.items():
            self.set(v, k)
    @staticmethod
    def dir(fp):
        return os.path.dirname(fp)
    @staticmethod
    def join(*a):
        return os.path.join(*a)
    @staticmethod
    def rjoin(path, *a):
        a = [k.strip() for k in a if k is not None and k.strip()!=""]
        if len(a)==0:
            return path
        if path is None:
            return os.path.join(*a)
        return os.path.join(path, *a)
    @staticmethod
    def rfp(paths, *a, last=-1, check_abs=False):
        a = [k.strip() for k in a if k is not None and k.strip()!=""]
        # if len(a)==0:
        #     a = ""
        if check_abs and len(a)>0:
            f = a[0]
            if f[:1]=="/" or f.find(":")>0:
                return Path.join(*a)
        #print(f"[TESTZ] a: {a}, paths: {paths}")
        #fp = os.path.join(*a)
        for path in paths:
            _fp = Path.rjoin(path, *a)
            if os.path.exists(_fp):
                return _fp
        return Path.rjoin(paths[last], *a)
    def add(self, path, name=None, index=0):
        self.paths[name].insert(index, path)
    def set(self, paths, name=None, last = -1, curr = None):
        if type(paths) not in (list, tuple):
            paths = [paths]
        if curr is not None and None not in paths:
            if Key.check_first(curr):
                paths = [None]+paths
            elif Key.check_last(curr):
                paths.append(None)
            else:
                assert False
        self.paths[name] = paths
        self.lasts[name] = last
        def fc(*a):
            return self.rfp(paths, *a, last=last, check_abs=self.check_abs)
        self.fcs[name] = fc
    def __getattr__(self, name):
        return self.fcs[name]
    def call(self, *obj):
        return self.fcs[None](*obj)

pass




from .. import xf, pyz, dz
from .sc import ScRunner
def test():
    import sys
    fp = sys.argv[1]
    conf = xf.loadf(fp)
    dz.s(conf, run="buildz.db.subrun.test")
    ScRunner.process_update(conf)

pass

pyz.lc(locals(), test)


#
import os
import sys

from buildz.fz import dirz as dirs
dirs.Deal = dirs.FileDeal
#from filez import dirs
import re
class Removes(dirs.Deal):
    def init(self):
        super(self.__class__, self).init()
        self.dps = []
    def result(self):
        self.dps.sort(key = lambda x:len(x),reverse=True)
        print("dps:", self.dps)
        for dp in self.dps:
            os.rmdir(dp)
        return self.dps
    def visit(self, fileinfo, depth):
        isdir = fileinfo.isdir
        filepath = fileinfo.path
        if not isdir:
            os.remove(filepath)
            return False
        self.dps.append(filepath)
        return True

pass
class Deal(dirs.Deal):
    def visit(self, fileinfo, depth):
        isdir = fileinfo.isdir
        filepath = fileinfo.path
        if not isdir:
            return False
        dp = filepath
        dirname = os.path.basename(dp)
        #if dirname[:1] == ".":
        #    return False
        if dirname not in ("__pycache__", "chromeCache"):
            return True
        Removes().dirs(dp)
        #os.removedirs(dp)
        return False
    def catch(self, fileinfo, depth, exp):
        isdir = fileinfo.isdir
        filepath = fileinfo.path
        print(f"exp in {filepath} isdir={isdir}: {exp}")
        pass

pass


def test():
    dp = sys.argv[1]
    Deal().dirs(dp)

pass

if __name__=="__main__":
    test()

pass




from .. import base
from .. import itemz
from ...readz import is_args
class ListMapDeal(base.BaseDeal):
    def height_tree(self, mg, obj):
        arr = obj.args
        maps = obj.maps
        rst = []
        height = 1
        rst_arr = []
        for obj in arr:
            _h = mg.height_tree(obj)
            height = max(height, _h[0]+1)
            rst_arr.append(_h)
        rst_maps = []
        for key, val in maps.items():
            h_key = mg.height_tree(key)
            h_val = mg.height_tree(val)
            height = max(height, h_key[0]+1, h_val[0]+1)
            rst_maps.append((h_key, h_val))
        return [height, rst_arr, rst_maps]
    def init(self, left, right, st, spt):
        self.left = left
        self.right = right
        self.st = st
        self.spt = spt
    def deal(self, mg, obj, up_height, heights, conf, deep, top_not_format):
        height = 1
        aft_set = conf.s(" ")*conf.get(set=0)
        s_set = conf.s(self.st)+aft_set
        arr = obj.args
        maps = obj.maps
        rst = []
        height = heights[0]
        for obj, _heights in zip(arr, heights[1]):
            sitem = mg.deal(obj, height, _heights, conf, deep+1)
            rst.append(sitem)
        for kv, _heights in zip(maps.items(), heights[2]):
            key, val = kv
            skey = mg.deal(key, height, _heights[0], conf, deep+1)
            sval = mg.deal(val, height, _heights[1], conf, deep+1, top_not_format = 1)
            sitem = skey+s_set+sval
            rst.append(sitem)
        return self.deal_encape(rst, conf, deep, height,up_height, self.left, self.right, top_not_format)

pass 



            

                


#
from .. import xf
from .. import ioc
from ..base import Base
from ..ioc import wrap
import os
@wrap.obj(id="buildz.auto.config.load")
@wrap.obj_set(cache="ref, cache")
class Config(Base):
    def call(self, maps, fp):
        configs = xf.g(maps, configs=[])
        if type(configs)!=list:
            configs = [configs]
        dts = {}
        for cfp in configs:
            cfp = self.cache.rfp(cfp)
            if not os.path.isfile(cfp):
                self.log.error(f"config file not exist: {cfp}")
                return False
            dt = xf.loadf(cfp)
            xf.fill(dt, dts, replace=1)
        xf.fill(dts, maps, replace=0)
        return True

pass
#wrap.add_datas("[(env, env.buildz.auto.deal), buildz.auto.deal, auto.deal]")
@wrap.obj(id="autoz.deal")
@wrap.obj_args("[env, buildz.auto.deal, auto.deal]")
class DfDeal(Base):
    def init(self, id):
        self.id = id
    def call(self, data):
        print(f"[ERROR] implement obj with id '{self.id}' by yourself")

pass






#coding=utf-8

__version__="0.0.1"

__author__ = "Zzz, emails: 1174534295@qq.com, 1309458652@qq.com"
# 小号多

from .middle_cache import MiddleCache

# 下面的dict_middle也弃了，上面的MiddleCache写的更合理
from .dict_middle import DictCache, Fcs
#from .dict_middle import *
Dict = DictCache

'''
20250917
写好的是MiddleCache
暂时别用DictCache，没测过

20251016
测试发现，数据在内存和显存间传输消耗太多，显存不足用此代码未必能提升多少性能，还不如加显存，或者把模型改小，不建议用
'''



from .base import *
from ... import dz,pyz
class EnvEncape(BaseEncape):
    def init(self, unit, env, default):
        super().init()
        self.unit = unit
        self.params = env, default
    def call(self, params=None, **_):
        env, default = self.params
        env = self.obj(env)
        obj, _, find = self.unit.get_env(env)
        if find:
            return obj
        assert default[0], f"env '{env}' not found"
        obj = default[1]
        obj = self.obj(obj)
        return obj
class EnvDeal(BaseDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.index(1, 'env', need=1)
        conf.index(2, 'default')
        conf.key('env', "data,key,profile,conf".split(","), need=1)
        self.update = conf
    def deal(self, conf, unit):
        assert 'env' in conf
        env = self.get_encape(conf['env'], unit)
        default, dfind = dz.dget(conf, 'default')
        if dfind:
            default = self.get_encape(default, unit)
        return EnvEncape(unit, env, [dfind, default])

pass


#coding=utf-8

from . import dirz

import re
class ListsDeal(dirz.FileDeal):
    def result(self):
        return self.fps, self.errors
    def init(self):
        super().init()
        self.fps = []
        self.errors = []
    def visit(self, finfo, depth):
        fp, isdir = finfo.path, finfo.isdir
        self.fps.append([fp, isdir])
        return True
    def catch(self, finfo, depth, exp):
        fp, isdir = finfo.path, finfo.isdir
        self.errors.append([fp, isdir, exp])

pass

def decode(s, coding = 'utf-8'):
    coding = coding.lower()
    xcoding = 'utf-8'
    if coding == 'utf-8':
        xcoding = 'gbk'
    try:
        return s.decode(coding)
    except:
        return s.decode(xcoding)

pass
g_codes = ["utf-8", "gbk"]
class SearchDeal(dirz.FileDeal):
    def init(self, pt_fp=None, pt = None, pt_dp = None, depth = None, relative = False, show_ct = False, shows = [10,10]):
        super().init()
        self.pt_fp = pt_fp
        self.pt_dp = pt_dp
        #if type(pt) == str:
        #    pt = pt.encode()
        if pt is not None:
            pfx = f"([\s\S]{{0,{shows[0]}}}"
            sfx = f"[\s\S]{{0,{shows[1]}}})"
            if type(pt)==bytes:
                pfx = pfx.encode()
                sfx = sfx.encode()
            pt = pfx+pt+sfx
        global g_codes
        if pt is not None:
            if type(pt) == str:
                pts = [pt.encode(c) for c in g_codes]
            else:
                pts = [pt]
        else:
            pts = None
        self.pts = pts
        self.pt = pt
        self.rst = []
        self.errs = []
        self.depth = depth
        self.relative = relative
        self.show_ct = show_ct
        self.show_list = []
    def fps(self, keep_dir = False, relative = None, fp_only = True):
        if self.show_ct:
            return self.show_list
        if relative is None:
            relative = self.relative
        rst = self.rst
        if relative:
            rst = [[i.rpath, i.isdir] for i in rst]
        else:
            rst = [[i.path, i.isdir] for i in rst]
        if not keep_dir:
            rst = [k for k in rst if not k[1]]
        if fp_only:
            rst = [k[0] for k in rst]
        return rst
    def result(self):
        if self.show_ct:
            return self.show_list
        return self.rst
    def visit(self, finfo, depth):
        if self.depth is not None and depth > self.depth:
            return False
        filepath = finfo.path
        isdir = finfo.isdir
        if self.relative:
            filepath = finfo.rpath
        if isdir:
            if finfo.empty_dir or depth==self.depth:
                if self.pt_dp is not None and len(re.findall(self.pt_dp, filepath))==0:
                    return True
                self.rst.append(finfo)
            return True
        if self.pt_fp is not None and len(re.findall(self.pt_fp, filepath))==0:
            return False
        if self.pt is not None:
            try:
                with open(filepath, 'rb') as f:
                    s = f.read()
            except Exception as exp:
                self.catch(finfo, depth, exp)
                return False
            pt = self.pt
            mark = False
            finds = None
            mark = -1
            for i in range(len(self.pts)):
                pt = self.pts[i]
                finds = re.findall(pt, s)
                if len(finds)>0:
                    mark = i
                    break
            if mark<0:
                return False
            if self.show_ct:
                arr = []
                code = g_codes[mark]
                bld = b"* "*32
                blda = b"- "*32
                bldi = b".."*32
                arr.append(bld)
                arr.append(f"filepath: {filepath}, find {len(finds)}:".encode("utf-8"))
                arr.append(bldi)
                if type(self.pt)==bytes:
                    pass
                else:
                    try:
                        s = s.decode(code)
                        pt = self.pts[mark]#.decode(code)
                        if type(pt)==bytes:
                            pt = pt.decode(code)
                        finds = re.findall(pt, s)
                    except Exception as exp:
                        print(f"error in decode {filepath}, {self.pts[mark]} with '{code}': {exp}")
                for bs_find in finds:
                    if type(bs_find)==[list, tuple]:
                        bs_find = bs_find[0]
                    if type(self.pt)==str:
                        if type(bs_find)==bytes:
                            bs_find = bs_find.decode(code)
                    #bs_find = b">>>>>>>\n"+bs_find+b"\n<<<<<<\n"
                    #bs_find += b"\n"+bld+b"\n"
                    arr.append(bs_find)
                    arr.append(bldi)
                arr.append(blda)
                arr.append(b"")
                arr =[k if type(k)==bytes else k.encode(code) for k in arr]
                arr = b"\n".join(arr)
                try:
                    arr = decode(arr)
                except Exception as exp:
                    arr = filepath
                self.show_list.append(arr)
        self.rst.append(finfo)
        return False
    def catch(self, finfo, depth, exp):
        filepath = finfo.path
        if self.relative:
            filepath = finfo.rpath
        print(f"exp in {finfo}: {exp}")
        self.errs.append([finfo, exp])
        pass

pass

def lists(fp):
    return ListsDeal().dirs(fp)

pass
def _search(dp, pt_fp = None, pt = None, pt_dp = None, depth=None, relative = False, show = False, shows = [10,10]):
    deal = SearchDeal(pt_fp, pt, pt_dp, depth,relative, show, shows)
    return deal.dirs(dp)

pass
def searchs(dp, pt_fp = None, pt = None, pt_dp = None, depth=None, relative = False, show = False, shows = [10,10]):
    deal = SearchDeal(pt_fp, pt, pt_dp, depth,relative, show, shows)
    deal.dirs(dp)
    return deal.fps(keep_dir=True, fp_only=False)

pass
def searchd(dp, pt_fp = None, pt = None, pt_dp = None, depth=None, relative = False, show = False, shows = [10,10]):
    deal = SearchDeal(pt_fp, pt, pt_dp, depth,relative, show, shows)
    deal.dirs(dp)
    return deal.fps(keep_dir=True, fp_only=True)

pass

def search(dp, pt_fp = None, pt = None, depth=None, relative = False, show = False, shows = [10,10]):
    deal = SearchDeal(pt_fp, pt, None, depth,relative, show, shows)
    deal.dirs(dp)
    return deal.fps(keep_dir=False, fp_only=True)

pass


#

import sys
from buildz.gpuz.torch import DictCache
from buildz.gpuz.test import analyze
from buildz import pyz
import torch,time
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
cpu,cuda = analyze.dvs
class ConvModel(nn.Module):
    def __init__(self, dims, ins_channels, middle_channels):
        super().__init__()
        nets=[]
        nets.append(nn.Conv2d(ins_channels, middle_channels, 5, padding=2))
        nets.append(nn.LeakyReLU())
        self.nets = nn.Sequential(*nets)
    def forward(self, inputs):
        return self.nets(inputs)
    def size(self):
        return sum([analyze.unit_sz(net) for net in self.nets])

pass
class UNet(nn.Module):
    def __init__(self, encoder, dealer, decoder):
        super().__init__()
        self.encoder, self.dealer, self.decoder = encoder, dealer, decoder
    def forward(self, inputs):
        tmp = self.encoder(inputs)
        out = tmp
        if self.dealer is not None:
            out = self.dealer(tmp)
        tmp = torch.cat([tmp, out], dim=1)
        return self.decoder(tmp)
class ResNet(nn.Module):
    def __init__(self, net):
        super().__init__()
        self.net = net
    def forward(self, inputs):
        return inputs+self.net(inputs)
class TestDataset(Dataset):
    def __init__(self, num, dims, channels, outs_dims):
        self.num = num
        self.dims = dims
        self.datas = torch.rand(num, channels, dims,dims)
        self.targets = torch.rand(num, outs_dims)
        sz = analyze.sz(self.datas)+analyze.sz(self.targets)
        sz, unit = analyze.show_size(sz)
        print(f"data size: {sz} {unit}")
    def __len__(self):
        return self.num
    def __getitem__(self, i):
        return self.datas[i], self.targets[i]

pass
class LnsNet(nn.Module):
    def __init__(self, input_dims, outs_dims, num):
        super().__init__()
        curr_dims = input_dims
        lns = []
        for i in range(num):
            rate = (i+1)/num
            dims = int(input_dims*(1-rate)+outs_dims*rate)
            ln = nn.Linear(curr_dims, dims)
            lns.append(ln)
            lns.append(nn.LeakyReLU())
            curr_dims = dims
        self.nets = nn.Sequential(*lns)
    def forward(self, inputs):
        batch = inputs.size(0)
        inputs = inputs.reshape(batch, -1)
        return self.nets(inputs)
    def size(self):
        return sum([analyze.unit_sz(net) for net in self.nets])

def gen(dims, outs_dims, channels, num_conv, num_ln, lr):
    """
        类似UNet的模型，不过cat改成了resnet的加
    """
    mds = []
    curr = channels
    encoders = []
    decoders = []
    for i in range(num_conv):
        encoder = ConvModel(dims, curr, curr*2)
        decoder = ConvModel(dims, curr*4, curr)
        curr*=2
        encoders.append(encoder)
        decoders.append(decoder)
    base = None
    mds += encoders
    mds+=decoders
    print(f"curr:{curr}")
    for i in range(num_conv):
        e = encoders[num_conv-i-1]
        d = decoders[num_conv-i-1]
        base = UNet(e, base, d)
    cvnets = base
    input_dims = dims*dims*channels
    lns = LnsNet(input_dims, outs_dims, num_ln)
    print(lns)
    mds.append(lns)
    gmds = [lns]
    mds_sz = [md.size() for md in mds]
    fullnet = nn.Sequential(cvnets, lns)
    opts =[optim.Adam(md.parameters(), lr=lr) for md in mds]
    gopt = optim.Adam(fullnet.parameters(), lr=lr)
    sz, unit = analyze.show_size(sum(mds_sz))
    print(f"Model Size: {sz} {unit}, mds: {len(mds)}")
    return mds, fullnet, opts, gopt, [gmds, None]
    


def fc_opt(net, opt):
    torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
    opt.step()
def test():
    channels=3
    dims= 32
    out_dims = 10
    loop = 10
    datas = 6
    batch=2
    lr=0.0001
    win_size=3
    num_conv = 6
    num_ln = 2
    args = sys.argv[1:]
    mark_train = True
    if len(args)>0:
        mark_train = args.pop(0).lower()=='train'
    modes = 'cuda,cache,cpu'
    if len(args)>0:
        modes = args.pop(0)
    if len(args)>0:
        num_conv = int(args.pop(0))
    if len(args)>0:
        num_ln = int(args.pop(0))
    print(f"num_conv: {num_conv}, num_ln: {num_ln}")
    ds = TestDataset(datas, dims, channels, out_dims)
    dl = DataLoader(ds, batch)
    loss_fn = torch.nn.MSELoss()
    def fc_gen():
        return gen(dims, out_dims, channels, num_conv, num_ln, lr)
    analyze.analyzes(mark_train, loop, fc_gen, dl, loss_fn, fc_opt, win_size, modes)

pyz.lc(locals(),test)




#print("xf.__init__")
from .readz import loads, load, loadf, loads_args, loadx, is_args, loadxf
from .writez import dumps, dump, dumpf, dumpx, dumpxf
from .file import *
from .mapz import *
from .xargs import fetch as args

__author__ = "Zzz, emails: 1174534295@qq.com, 1309458652@qq.com"



#

from .lst import FcFpsListener
from .. import Base, xf, dz, pathz, pyz
import os, sys
class Runner(Base):
    def init(self, fp, lst):
        fp = os.path.abspath(fp)
        self.dp = os.path.dirname(fp)
        self.path = pathz.Path(fp=self.dp)
        self.fp = fp
        self.lst = lst
        self.lst.set_update(self.update)
        self.lst.set_deal_exp(self.deal_exp)
        self.reset()
    def reset(self,last_update=False):
        conf = xf.loadf(self.fp)
        fps, target, dp = dz.g(conf, fps = [], run=None, dp="")
        dp = self.path.fp(dp)
        sys.path.append(dp)
        self.path.set("sc", dp)
        fps = [self.path.sc(fp) for fp in fps]
        self.lst.clean()
        self.lst.add(self.fp, last_update)
        [self.lst.add(fp, last_update) for fp in fps]
        self.target=target
    def update(self, fps):
        self.reset(True)
        fc = pyz.load(self.target)
        fc()
    def deal_exp(self, exp, fmt_exc):
        print(f"exp: {exp}")
        print(f"traceback: \n{fmt_exc}")

def test():
    fp = sys.argv[1]
    lst = FcFpsListener()
    runner = Runner(fp, lst)
    lst.run()

pass
pyz.lc(locals(), test)


#coding=utf-8
from buildz import xf, argx, pyz, ioc

from ..test import text_sfx, args, maps
class Deal:
    def deal(self):
        if len(args)<2:
            print("need params 1 to be filepath")
            return
        fp = args[1]
        obj = xf.loads(xf.fread(fp))
        s = xf.dumps(obj, format=1, deep=1)
        print(s)
    pass

pass



#

from .base import *
from .encapes import Encapeset
from .dataset import Dataset
from .vars import Varset
from ... import pyz,dz
from .ids import Ids
from .unit import Unit
from .builds import Buildset
from .sys_envs import SysEnvs
class GetKey(Base):
    def call(self, conf):
        return None, 0
    def fill(self, conf, key):
        assert 0,'not callable'
class SimpleConfKey(Base):
    def init(self, key):
        self.key = key
    def call(self, conf):
        if self.key not in conf:
            return None,0
        return conf[self.key],1
    def fill(self, conf, key):
        conf[self.key] = key

pass
class EnvKey:
    sys = 'sys'
    args = 'args'
    conf = 'conf'
    local = 'local'
    default_orders = [local, args, conf, sys]
class Manager(Base):
    EnvKey = EnvKey
    @staticmethod
    def make_key(key):
        if type(key) == str:
            key = SimpleConfKey(key)
        return key
    @staticmethod
    def make_ids(ids):
        if type(ids)==str:
            ids = Ids(ids)
        return ids
    def get_env(self, key, ns = None, tag = None, id = None, local_get = None):
        fcs = self.get_env_fcs
        fcs[EnvKey.local] = local_get
        for order in self.env_orders:
            fc = fcs[order]
            if fc is None:
                continue
            obj, _tag, find = fc(key, ns, tag, id)
            if find:
                return obj, _tag, find
        return None, None, 0
    def set_env_args(self, key, val, ns=None):
        self.sys_env.set(key, val, ns)
    def init(self, ids, deal_key='type', conf_key='id', tag_key='tag', deal_ids = None, env_ids = None, env_orders = None):
        if env_orders is None:
            env_orders = EnvKey.default_orders
        self.env_orders = env_orders
        self._index = 0
        self.units = {}
        self.builds = Buildset(self)
        ids = self.make_ids(ids)
        self.ids = ids
        self.deal_key = self.make_key(deal_key)
        self.conf_key = self.make_key(conf_key)
        self.tag_key = self.make_key(tag_key)
        self.deal_ids = self.make_ids(pyz.nnull(deal_ids, ids))
        self.env_ids = self.make_ids(pyz.nnull(env_ids, ids))
        self.confs = Dataset(self.ids)
        self.envs = Dataset(self.env_ids)
        self.sys_envs = SysEnvs(self.env_ids)
        self.args_envs = Dataset(self.env_ids)
        self.deals = Dataset(self.deal_ids)
        self.encapes = Encapeset(self.ids, self)
        self.vars = Varset(self.ids)
        #
        self.set_env_args = self.args_envs.set
        self.update_env_args = self.args_envs.update
        self.update_env = self.envs.update
        self.set_env = self.envs.set
        #self.get_env = self.envs.tget
        self.get_env_fcs = {
            EnvKey.sys: self.sys_envs.tget,
            EnvKey.args: self.args_envs.tget,
            EnvKey.conf: self.envs.tget
        }
        #
        self.push_var = self.vars.vpush
        self.push_vars = self.vars.vpushs
        self.pop_var = self.vars.vpop
        self.pop_vars = self.vars.vpops
        self.set_var = self.vars.vset
        self.set_vars = self.vars.vsets
        self.get_var = self.vars.vget
        self.unset_var = self.vars.vremove
        self.unset_vars = self.vars.vremoves
        #
        self.default_unit = self.create()
    def add_build(self, conf):
        self.builds.add(conf)
    def build(self):
        self.builds.build()
    def add(self, unit):
        unit.bind(self)
        self.units[unit.id] = unit
    def create(self, ns=None, deal_ns = None, env_ns = None, deal_key=None, conf_key= None):
        id = self.id()
        deal_key = pyz.nnull(deal_key, self.deal_key)
        conf_key = pyz.nnull(conf_key, self.conf_key)
        deal_key = self.make_key(deal_key)
        conf_key = self.make_key(conf_key)
        unit = Unit(ns, deal_ns, env_ns, deal_key, conf_key, id)
        unit.bind(self)
        return unit
    def get_unit(self, id):
        if id not in self.units:
            return self.default_unit
        return self.units[id]
    def id(self):
        index = self._index
        self._index+=1
        return index
    def get_deal(self, key, ns=None, id=None):
        return self.deals.tget(key, ns, id)
    def set_deal(self, key, deal, ns=None, tag=None, id=None):
        self.deals.set(key, deal, ns, tag, id)
    def get_conf(self, key, ns=None, id=None):
        return self.confs.tget(key, ns, id)
    def set_conf(self, key, conf, ns=None, tag=None, id=None):
        self.confs.set(key, conf, ns, tag, id)
    def get_encape(self, key, ns=None, id=None):
        self.build()
        return self.encapes.tget(key, ns, id)
    def set_encape(self, key, encape, ns=None, tag=None, id=None):
        self.encapes.set(key, encape, ns, tag, id)
    def get(self, key, ns=None, params=None, id=None, search_var = True):
        if search_var:
            obj, find = self.get_var(key, ns)
            if find:
                return obj, 1
        encape, tag, find = self.get_encape(key, ns, id)
        if not find:
            return None, 0
        return encape(params),1
    def get_obj(self, key, ns=None, params=None, id=None, search_var=True):
        return self.get(key, ns, params, id, search_var)[0]



pass



from ..base import Base
class Lists(Base):
    def init(self, *keys):
        self.lists = {}
        for k in keys:
            self.lists[k] = []
        self.enables = set()
    def add(self, kid, _type, fc, enable=True):
        if _type:
            if _type not in self.lists:
                self.lists[_type] = []
            self.lists[_type].append([kid, fc])
        else:
            for _type in slef.lists.keys():
                self.lists[_type].append([kid, fc])
        self.enable(kid, enable)
    def adds(self, kid, enable, **maps):
        for _type, fc in maps.items():
            self.add(kid, _type, fc, enable)
    def enable(self, kid, val=True):
        if val:
            self.enables.add(kid)
        elif kid in self.enables:
            self.enables.remove(kid)
    def disable(self, kid):
        return self.enable(kid, False)
    def call(self, _type, data, *a, **b):
        for kid, fc in self.lists[_type]:
            if kid not in self.enables:
                continue
            data = fc(data, *a, **b)
        return data
    def __getattr__(self, key):
        if key in {'lists', 'enables'}:
            return super().__getattr__(key)
        def fc(data, *a, **b):
            return self.call(key, data, *a, **b)
        return fc


    




#coding=utf-8
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from cryptography.x509.oid import NameOID
import datetime
from buildz import xf, pyz
from buildz import fz
from .pk import *
from . import gen_names
__doc__='''
cert和csr的区别：
    csr是用自己的私钥签名，对象里包含自身信息和公钥（用公钥可以验证私钥签名是否有效），用于发给证书机构，让证书机构验证和签发cert用的
    cert是被证书机构私钥签名，包含被签对象自身信息和公钥，以及签名机构的信息（无公钥），对外使用
        cert验证过程:
            根据签名机构信息查找签名机构的证书（验证方要自己存储签名机构公钥）
    cert链：
        被签对象的cert，中间节点的cert，CA(根签名机构)的cert（自签名，可以不包含）
        验证过程:
            用上层节点的cert的公钥验证下层节点cert是否是它签名的，
            最后用CA的cert的公钥验证它自己（为了安全，验证方要自己存储CA的自签名cert证书，用自己存的cert的公钥验证cert链的最高层）
'''
arr_names = gen_names.loadf()
dict_names = {k[0]:k for k in arr_names}
def a2n(arr_names):
    rst = {}
    for item in arr_names:
        oid = item[0]
        als = item[3]
        rst[oid] = oid
        for al in als:
            rst[al] = oid
    return rst
al2names = a2n(arr_names)
def get_subject_val(obj, key):
    '从subject（或cert/csr）获取key对应的值，返回列表'
    if key not in al2names:
        return []
    key = al2names[key]
    key = getattr(NameOID, key)
    if not hasattr(obj, 'get_attributes_for_oid') and hasattr(obj, 'subject'):
        obj = obj.subject
    rst = obj.get_attributes_for_oid(key)
    rst = [k.value for k in rst]
    return rst
get_sub_val = get_subject_val
def gen_subject(conf = {}):
    "生成证书的主体信息"
    names = []
    for item in arr_names:
        OID,des,df,als = item
        val = conf.get(OID)
        for al in als:
            if val:
                break
            val = conf.get(al)
        val = val or df
        if val:
            names.append(x509.NameAttribute(getattr(NameOID, OID), val))
    subject = x509.Name(names)
    return subject

pass
def add_extension_dns(builder, dns):
    '添加dns信息，暂不清楚有什么用'
    if type(dns)!=list:
        dns = [dns]    
    dns_arr = [x509.DNSName(k) for k in dns]
    dns_arr = x509.SubjectAlternativeName(dns_arr)
    builder = builder.add_extension(dns_arr, critical=False)
    return builder
def add_extension_ca(builder):
    '添加ca信息（说明当前证书是ca）'
    builder = builder.add_extension(
        x509.BasicConstraints(ca=True, path_length=None),
        critical=True
    )
    return builder
def add_extensions(builder, conf = {}):
    "添加额外信息"
    dns = xf.g(conf, dns = [])
    if len(dns)>0:
        builder = add_extension_dns(builder, dns)
    ca = xf.g(conf, ca = False)
    if ca:
        builder = add_extension_ca(builder)
    return builder

pass

def load_csr(pem):
    '从bytes载入csr对象'
    csr= x509.load_pem_x509_csr(pem)
    return csr 
pass
def load_cert(pem):
    '从bytes载入cert对象'
    cert = x509.load_pem_x509_certificate(pem)
    return cert

def gen_cert(private_key, subject_csr = None, issuer_cert = None, conf={}, to_pem=True, add_chains=True):
    '''
       生成cert或csr
       to_pem=True则返回序列化后的pem的bytes数据
       否则返回cert或csr对象
       private_key: 私钥对象，必须是对象
       subject_csr: 可选值:
                        csr对象或序列化bytes
                        空
                    为空的时候，从conf生成csr对象
       issuer_cert: 可选值:
                        用于签名的cert对象或序列化bytes
                        布尔值(True/False/1/0)
                        空
                    issuer_cert为布尔值True/1的时候，将subject_csr当作cert，也就是自签名
                    issuer_cert为布尔值False/0或空的时候，方法返回csr，否则返回cert
       conf: 配置信息，用于生成csr和cert签名的时候设置日期属性
       to_pem: 是否返回序列化后的bytes
       add_chains: 是否返回证书链，只在to_pem=True, issuer_cert非空并且非布尔值False/0时生效
    '''
    if issuer_cert:
        builder = x509.CertificateBuilder()
    else:
        builder = x509.CertificateSigningRequestBuilder()
    if subject_csr is not None:
        if type(subject_csr)==bytes:
            subject_csr = load_csr(subject_csr)
        subject = subject_csr.subject
        verify_csr(subject_csr)
    else:
        subject = gen_subject(conf)
    builder = builder.subject_name(subject)
    issuer = None
    issuer_src = None
    if issuer_cert:
        if type(issuer_cert) in {bool, int}:
            issuer = subject
        else:
            issuer_src = issuer_cert
            if type(issuer_cert)==bytes:
                issuer_cert = load_cert(issuer_cert)
            issuer = issuer_cert.subject
    if issuer:
        builder = builder.issuer_name(issuer)
        if subject_csr:
            public_key=subject_csr.public_key()
        else:
            public_key = private_key.public_key()
        builder = builder.public_key(public_key)
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.not_valid_before(datetime.datetime.utcnow() - datetime.timedelta(days=xf.g(conf, valid_before=1)))
        builder = builder.not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=xf.g(conf, valid=365)))
    if subject_csr:
        for extension in subject_csr.extensions:
            builder = builder.add_extension(extension.value, extension.critical)
    else:
        builder = add_extensions(builder, conf)
    # 将证书签名
    certificate = builder.sign(
        private_key=private_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )
    if to_pem:
        certificate = cert2pem(certificate, issuer_src, add_chains)
        #certificate = certificate.public_bytes(serialization.Encoding.PEM)
        #if add_chains and issuer_src:
        #    if type(issuer_src)!=bytes:
        #        issuer_src = issuer_src.public_bytes(serialization.Encoding.PEM)
        #    certificate += issuer_src
    return certificate
pass
def cert2pem(cert, up=None, add_chains=True):
    issuer_src=up
    certificate = cert.public_bytes(serialization.Encoding.PEM)
    if add_chains and issuer_src:
        if type(issuer_src)!=bytes:
            issuer_src = issuer_src.public_bytes(serialization.Encoding.PEM)
        certificate += issuer_src
    return certificate
def gen_csr(private_key, conf={}, to_pem=True):
    return gen_cert(private_key, conf=conf, to_pem=to_pem)
pass
def sign_csr(private_key, subject_csr, issuer_cert, conf={}, to_pem=True, add_chains=True):
    assert subject_csr is not None and issuer_cert is not None
    return gen_cert(private_key, subject_csr, issuer_cert, conf, to_pem, add_chains)
def gen_root_cert(private_key, conf={}, to_pem=True):
    return gen_cert(private_key, issuer_cert=True, conf=conf, to_pem=to_pem)
#要生成证书链，是从网站证书开始写入，往上写到根证书
"""
chains.cert文件结构:
网站证书
中间节点证书
...
根证书

该结构保证load_cert(数据)的时候，拿到的是最上面的网站证书
"""
class CertVerifyException(Exception):
    pass

pass
def load_certs(bs):
    '加载证书对象列表'
    spt = b"-----END CERTIFICATE-----"
    pems = bs.split(spt)
    pems = [k+spt for k in pems if k.strip()!=b""]
    certs = [load_cert(pem) for pem in pems]
    return certs
def verify_certs(certs, cas = None, verify_time=True):
    if type(certs)==bytes:
        certs = load_certs(certs)
    if certs is None:
        certs = []
    if type(certs)!=list:
        certs = [certs]
    if len(certs)==0:
        raise CertVerifyException(f"not any cert info in certs")
    if cas is None:
        cas = []
    if type(cas)!=list:
        cas = [cas]
    #cas = [load_cert(ca) for ca in cas]
    roots = {ca.subject:[ca.public_key(), ca.signature_hash_algorithm] for ca in cas}
    #print(f"[TESTZ] verify certs: {len(certs)}")
    for i in range(len(certs)-1):
        #print(f"verify[{i}]")
        curr = certs[i]
        up = certs[i+1]
        err = verify_cert(curr, up.public_key(), verify_time, up.signature_hash_algorithm)
        if err:
            return err
    #print(f"verify last")
    cert = certs[-1]
    issuer = cert.issuer
    if issuer == cert.subject and len(roots)==0:
        roots[cert.subject] = [cert.public_key(), cert.signature_hash_algorithm]
    if issuer not in roots:
        return f"not root cert find: {issuer}"
        raise CertVerifyException(f"not root cert find: {issuer}")
    pk, hash_alg = roots[issuer]
    return verify_cert(cert, pk, verify_time, hash_alg)

pass

def verify_csr(csr, public_key = None, hash_alg=None):
    if type(csr)==bytes:
        csr = load_csr(csr)
    return verify_cert(csr, public_key, False, hash_alg, csr=True)
def verify_cert(cert, public_key=None, verify_time=True, hash_alg = None, csr=False):
    '''
        证书验证，用public_key验证证书，如果public_key为空，用证书的public_key验证（只有自签名证书才能用证书的公钥验证）
        verify_time: 是否验证证书是否在有效期（证书签名的时候可以设置有效期，默认是1年有效期）
        验证通过返回空，否则返回报错信息
    '''
    if type(cert)==bytes:
        cert = load_cert(cert)
    if public_key is None:
        public_key = cert.public_key()
    if hash_alg is None:
        hash_alg = cert.signature_hash_algorithm
    if csr:
        tbs_bytes = cert.tbs_certrequest_bytes
    else:
        tbs_bytes = cert.tbs_certificate_bytes
    try:
        public_key.verify(cert.signature,tbs_bytes, padding.PKCS1v15(),hash_alg)
    except Exception as exp:
        return f"sign verify error: {exp}"
    if not verify_time:
        return None
    if cert.not_valid_before_utc<datetime.datetime.now(datetime.timezone.utc)<cert.not_valid_after_utc:
        return None
    return "date not pass"
    raise CertVerifyException("date not pass")

pass

def load_der(der):
    cert = x509.load_der_x509_certificate(der)
    return cert

pass
def mk_load(fc):
    def wrap(fp, *a, **b):
        bs = fz.read(fp)
        return fc(bs, *a, **b)
    return wrap
def mk_gen(fc):
    def wrap(fp, *a, **b):
        bs = fc(*a, **b)
        fz.write(bs, fp)
    return wrap
loadf_cert = mk_load(load_cert)
loadf_csr = mk_load(load_csr)
loadf_certs = mk_load(load_certs)
genf_cert = mk_gen(gen_cert)
genf_csr = mk_gen(gen_csr)
signf_csr = mk_gen(sign_csr)
genf_root_cert = mk_gen(gen_root_cert)
verifyf_certs = mk_load(verify_certs)
verifyf_cert = mk_load(verify_cert)
verifyf_csr = mk_load(verify_csr)
loadf_prv=mk_load(load_prv)
loadf_pub = mk_load(load_pub)
genf_prv = mk_gen(gen_prv)
genf_pub = mk_gen(gen_pub)

SOIDS = "COUNTRY_NAME,STATE_OR_PROVINCE_NAME,LOCALITY_NAME,ORGANIZATION_NAME,COMMON_NAME,EMAIL_ADDRESS".split(",")
OIDS = ['BUSINESS_CATEGORY', 'COMMON_NAME', 'COUNTRY_NAME', 'DN_QUALIFIER', 'DOMAIN_COMPONENT', 'EMAIL_ADDRESS', 'GENERATION_QUALIFIER', 'GIVEN_NAME', 'INITIALS', 'INN', 'JURISDICTION_COUNTRY_NAME', 'JURISDICTION_LOCALITY_NAME', 'JURISDICTION_STATE_OR_PROVINCE_NAME', 'LOCALITY_NAME', 'OGRN', 'ORGANIZATIONAL_UNIT_NAME', 'ORGANIZATION_IDENTIFIER', 'ORGANIZATION_NAME', 'POSTAL_ADDRESS', 'POSTAL_CODE', 'PSEUDONYM', 'SERIAL_NUMBER', 'SNILS', 'STATE_OR_PROVINCE_NAME', 'STREET_ADDRESS', 'SURNAME', 'TITLE', 'UNSTRUCTURED_NAME', 'USER_ID', 'X500_UNIQUE_IDENTIFIER']
#oids = [getattr(NameOID, oid) for oid in OIDS if hasattr(NameOID, oid)]

NAMES = ['业务类别', '公共名称', '国家名称', ' DN_限定符', '域_组件', '电子邮件_地址', '代_限定符', '名字', '旅馆', '管辖区_国家名称', '管辖区_地点_名称', '管辖区_州_或_省_名称', '地点_名称', 'OGRN', '组织_单位_名称', '组织_标识符', '组织_名称', '邮政_地址', '邮政_代码', '假名', '序列号', ' SNILS ', '州_或_省_名称', '街道_地址', '姓氏', '头衔', '非结构化名称', '用户_ID ', 'X500']
names = {getattr(NameOID, OID): f"{NAME}({OID.lower()})" for NAME, OID in zip(NAMES, OIDS) if hasattr(NameOID, OID)}
def des_subject(subject):
    arr = subject._attributes
    rst = {}
    for its in arr:
        for it in its:
            oid = it.oid
            val = it.value
            if oid not in dict_names:
                name = oid.dotted_string
            else:
                name = f"{dict_names[oid][1]}({dict_names[oid][0].lower()})"
            rst[name]=val
    return rst

pass
def test():
    pass

pass

if __name__=="__main__":
    test()

pass



#coding=utf-8
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from cryptography.x509.oid import NameOID
import datetime
from buildz import xf
from buildz import fz
def sign(private_key, bs):
    '私钥签名，只返回签名数据，不包括本体'
    return private_key.sign(bs, padding.PKCS1v15(), hashes.SHA256())
def verify(public_key, bs, sig):
    '公钥认证'
    public_key.verify(sig,bs, padding.PKCS1v15(), hashes.SHA256())

pass
def encrypt(public_key, bs):
    '公钥加密'
    encrypted = public_key.encrypt(
        bs,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return encrypted

pass
def decrypt(private_key, bs):
    '私钥解密'
    decrypted = private_key.decrypt(
        bs,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return decrypted

pass
# 生成RSA私钥
def gen_prv(pwd = None, to_pem=True):
    """
        生成pem格式私钥
    """
    if type(pwd)==str:
        pwd = pwd.encode()
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )
    if not to_pem:
        return private_key
    # 将私钥序列化为PEM格式
    pem = prv2pem(private_key, pwd)
    return pem
    if pwd is None:
        alg = serialization.NoEncryption()
    else:
        alg = serialization.BestAvailableEncryption(pwd)
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=alg
    )
    #return pem

pass
def gen_pub(private_key, pwd = None, to_pem=True):
    """
        从私钥生成pem格式公钥
    """
    if type(private_key)==bytes:
        private_key = load_prv(private_key, pwd)
    # 生成RSA公钥
    public_key = private_key.public_key()
    if not to_pem:
        return public_key
    # 将RSA公钥序列化为PEM格式
    pem = pub2pem(public_key)
    return pem
    pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    #return pem

pass

def load_prv(pem_data, pwd = None):
    if type(pwd)==str:
        pwd = pwd.encode()
    '将PEM格式的私钥反序列化为私钥对象'
    private_key = serialization.load_pem_private_key(pem_data, password=pwd, backend=default_backend())
    return private_key

pass
def load_pub(pem_data):
    '将PEM格式的公钥反序列化为公钥对象'
    public_key = serialization.load_pem_public_key(pem_data, backend=default_backend())
    return public_key

pass

def prv2pem(private_key, pwd=None):
    if pwd is None:
        alg = serialization.NoEncryption()
    else:
        alg = serialization.BestAvailableEncryption(pwd)
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=alg
    )
    return pem

def pub2pem(public_key):
    pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    return pem



#coding=utf-8
from ..ioc.base import Base, EncapeData,IOCError
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class DealDeal(FormatDeal):
    """
        扩展的好处是自定义格式，除了id,type和parent字段，其他字段可以自己加
        假设生成个test.Demo，构造函数只有一个字段val
        {
            id: test
            type: object
            source: test.Demo
            construct: {
                args: [
                    (val, 'demo')
                ]
            }
        }
        自己写个生成特定对象的处理类:
        配置文件1 def_demo.js:
        {
            datas:[
                {
                    id: obj.demo.deal
                    type: object
                    source: test.BuildDemo
                }
                {
                    id: demo.deal
                    type: deal
                    source: obj.demo.deal
                    target: demo.deal
                }
            ]
            //inits初始化
            inits: [
                demo.deal
            ]
        }
        配置文件2 demo.js:
        //只有datas的时候可以简写成只写datas列表里的数据
        [
            {
                id: test
                type: demo.deal
                val: "demo"
            }
        ]
        调用后会注册到conf的deal上，用于扩展deals配置
        deal字段deal:
            {
                id:id
                type: deal
                target: type
                source: id # 要求source实现了方法__call__(self, edata:EncapeData)
            }
        简写:
            [[deal, id], target, source]
            [deal, target, source]
        例:
            [deal, target, source] //
    """
    def init(self, fp_lists=None, fp_defaults=None):
        super().init("DealDeal", fp_lists, fp_defaults, join(dp, "conf", "deal_lists.js"), None)
    def deal(self, edata:EncapeData):
        data = edata.data
        data = self.fill(data)
        source = xf.g(data, source=None)
        if source is None:
            raise IOCError("not source in dealdeal")
        target = xf.g(data, target=None)
        if target is None:
            raise IOCError("not target in dealdeal")
        obj = edata.conf.get(source)
        if obj is None:
            raise IOCError("source object not found in dealdeal")
        targets = target
        if type(targets) != list:
            targets = [targets]
        for target in targets:
            edata.conf.set_deal(target, obj)
        return None

pass



#
from .build import * 

class PodmanBuilder(Builder):
    def has_image(self, s):
        if s == 'scratch' or s == 'local/scratch':
            return True
        _s = "localhost/"+s
        rst = cmd('podman images --format "{{.Repository}}:{{.Tag}}"|grep "'+s+'"').strip()
        return rst == s or rst == _s
    def has_container(self, s, tag=None):
        '''
            判断容器是否存在
        '''
        if tag is None:
            cmp = s
            cmpx = cmp
            s = 'podman ps -a --format "table {{.Names}}"|grep "'+s+'"'
        else:
            cmp = f"{s}:{tag}"
            cmpx = f"{s}:localhost/{tag}"
            s = 'podman ps -a --format "table {{.Names}}:{{.Image}}"|grep "'+cmp+'"'
        rst = cmd(s).strip()
        return rst == cmp or rst == cmpx
    def default_test(self, tag):
        return f"podman run --userns=keep-id -it --rm {tag}"
    def build_cmd(self, fp, tag, dp, s_args=""):
        return f"podman build -f {fp} -t {tag} {dp} && podman image prune -f"

if __name__=="__main__":
    exit(PodmanBuilder().demo())

pass



#
from .datas import Datas
from .dataset import Dataset
from .confs import Confs
class Envs(Datas):
    def init(self, ns=None, id=None, dts=None):
        super().init(ns, id, dts)
        self.ids = None
        self.caches = {}
    def update(self, maps, tag=None, flush=False):
        if flush:
            maps = dz.unflush_maps(maps, self.ids.id)
        for key, val in maps.items():
            self.set(key, val, tag)
    def bind(self, dts):
        self.ids = dts.ids
        super().bind(dts)
        for k,v in self.caches.items():
            val,tag=v
            self.set(k,val,tag)
        self.caches = {}
    def set(self, key, val, tag=None):
        if self.ids is not None:
            super().set(self.ids(key), val, tag)
        else:
            self.caches[key] = val,tag
    def get(self, key, tags=None):
        return super().get(self.ids(key), tags)



from cryptography.x509.oid import NameOID
OIDS = ['BUSINESS_CATEGORY', 'COMMON_NAME', 'COUNTRY_NAME', 'DN_QUALIFIER', 'DOMAIN_COMPONENT', 'EMAIL_ADDRESS', 
        'GENERATION_QUALIFIER', 'GIVEN_NAME', 'INITIALS', 'INN', 'JURISDICTION_COUNTRY_NAME', 'JURISDICTION_LOCALITY_NAME', 
        'JURISDICTION_STATE_OR_PROVINCE_NAME', 'LOCALITY_NAME', 'OGRN', 'ORGANIZATIONAL_UNIT_NAME', 'ORGANIZATION_IDENTIFIER', 'ORGANIZATION_NAME', 
        'POSTAL_ADDRESS', 'POSTAL_CODE', 'PSEUDONYM', 'SERIAL_NUMBER', 'SNILS', 'STATE_OR_PROVINCE_NAME', 
        'STREET_ADDRESS', 'SURNAME', 'TITLE', 'UNSTRUCTURED_NAME', 'USER_ID', 'X500_UNIQUE_IDENTIFIER']
NAMES = ['业务类别', '公共名称', '国家名称', ' DN_限定符', '域_组件', '电子邮件_地址', 
         '代_限定符', '名字', '旅馆', '管辖区_国家名称', '管辖区_地点_名称', '管辖区_州_或_省_名称', 
         '地点_名称', 'OGRN', '组织_单位_名称', '组织_标识符', '组织_名称', '邮政_地址', 
         '邮政_代码', '假名', '序列号', ' SNILS ', '州_或_省_名称', '街道_地址', 
         '姓氏', '头衔', '非结构化名称', '用户_ID ', 'X500']
from ... import xf
from os.path import join, dirname, isfile
dp = dirname(__file__)
def gen():
    als = xf.loads(r"""
    COUNTRY_NAME: [CN,country]
    STATE_OR_PROVINCE_NAME: [CN, provice],
    LOCALITY_NAME: [cityz, local],
    ORGANIZATION_NAME: [orgz,org],
    COMMON_NAME: [commonz, common],
    EMAIL_ADDRESS: [emailz,email]
    """)
    rst = []
    for OID, NAME in zip(OIDS, NAMES):
        df = ""
        obj = getattr(NameOID, OID)
        an=[obj.dotted_string]
        if obj._name!="Unknown OID":
            an.append(obj._name)
        al = als.get(OID)
        if al is not None:
            df = al[0]
            an.append(al[1])
        an = " | ".join(an)
        tmp = f"{OID}, {NAME}, {df}, {an}"
        rst.append(tmp)
    rs = "\n".join(rst)
    rs = "NAMEOID, Des, Default, Aliase\n"+rs
    return rs
def loadf(fp=None):
    fp = fp or join(dp,"names.txt")
    if not isfile(fp):
        write(fp)
    with open(fp, 'rb') as f:
        s = f.read().decode("utf-8")
    return load(s)

def load(s=None):
    s = s or gen()
    arr = s.split("\n")[1:]
    arr = [k.split(",") for k in arr]
    arr = [[v.strip() for v in k] for k in arr]
    for it in arr:
        it[-1] = it[-1].split("|")
        it[-1] = [k.strip() for k in it[-1] if k.strip()!=""]
    return arr

pass
def write(fp=None):
    fp = fp or join(dp,"names.txt")
    rs = gen()
    with open(fp, 'wb') as f:
        f.write(rs.encode("utf-8"))




from .base import Base
class Ids(Base):
    '''
    '''
    def init(self, spt):
        self.spt = spt
    def id(self, *ids):
        if len(ids)==0:
            return None
        elif len(ids)==1:
            if type(ids[0]) in (list, tuple):
                ids = ids[0]
        return self.spt.join(ids)
    def ids(self, id):
        if id is None:
            return []
        if type(id) in (list, tuple):
            return id
        return id.split(self.spt)
    def call(self, id):
        return self.ids(id)
pass




#

from . import obj
from . import env
def build(wraps):
    wraps.set_fc("obj", obj.ObjectWrap())
    fc = env.EnvWrap()
    wraps.set_fc("env", fc)
    wraps.set_fc("load_envs", fc)
    wraps.set_fc("profile", fc)
    wraps.set_fc("load_profiles", fc)
    return wraps

pass



import threading,time
from .. import Base
class Task(Base):
    def run(self):
        pass
    def deal_exp(self, exp, traceback_exp):
        pass
class BasePool(Base):
    def check_run(self):
        return False
    def get_task(self):
        return None
class Worker(Base):
    def id(self):
        return threading.get_ident()
    def init(self, pool):
        self.pool = pool
        self.running = False
        self.do_run = False
        self.wait_time=0
        self.wait_count = 0
        self.work_time=0
        self.th = None
    def start(self):
        assert self.th is None
        self.th = threading.Thread(target=self.run,daemon=True)
        self.th.start()
    def safe_stop(self):
        self.do_run=False
    def work_rate(self):
        return self.work_time/(self.work_time+self.wait_time+1e-4)
    def run(self):
        self.wait_time=0
        self.work_time=0
        self.wait_count=0
        self.running = True
        self.do_run = True
        while self.do_run and self.pool.check_run():
            start = time.time()
            task = self.pool.get_task()
            task_start = time.time()
            if task is None:
                self.wait_time+=task_start-start
                self.wait_count+=1
                continue
            self.wait_count=0
            try:
                task.run()
            except Exception as exp:
                import traceback
                try:
                    task.deal_exp(exp, traceback.format_exc())
                except Exception as _exp:
                    print(f"deal_exp on {exp} get except: {_exp}")
            finish = time.time()
            self.wait_time+=task_start-start
            self.work_time+=finish-task_start
        self.running = False
        self.th = None

pass
class Tasks(Base):
    def waits(self):
        return 0
    def add(self, task):
        pass
    def pop(self):
        return None
class WorkerManager(Base):
    def change(self, workers, tasks, first_call=False):
        return 0

pass
class Pool(BasePool):
    def check_run(self):
        return self.running
    def get_task(self):
        return self.tasks.pop()
    def init(self, tasks, manager):
        self.running = False
        self.workers = []
        self.stops = []
        self.manager = manager
        self.tasks = tasks
    def start(self):
        if self.running:
            return
        self.running = True
        for wk in self.workers:
            wk.safe_stop()
            self.stops.append(wk)
        self.workers = []
        self.clean_stops()
        self.change_workers(True)
    def change_workers(self, first=False):
        num = self.manager.change(self.workers, self.tasks, first)
        if num==0:
            return
        if num>0:
            for i in range(num):
                worker = Worker(self)
                worker.start()
                self.workers.append(worker)
        else:
            for i in range(-num):
                worker = self.workers.pop()
                worker.safe_stop()
                self.stops.append(worker)
    def clean_stops(self):
        self.stops = [wk for wk in self.stops if wk.running]
    def check_finish(self):
        for k in self.workers:
            if k.running:
                return False
        for k in self.stops:
            if k.running:
                return False
        return True
    def stop(self):
        self.running = False
    def add_task(self, task):
        self.tasks.add(task)
        if self.running:
            self.change_workers()

pass


#coding=utf-8
#author:
from .dirz import dirs, FileDeal
from .lsf import lists, search, searchd,searchs,_search
from .fio import fread, freads, fwrite, fwrites, read,reads,write,writes, makefdir, dirpath,makedir,removes, cover, fcover, sread, swrite, is_abs, xors, fxors, coverx, covers, fcovers
from .fhs import fhash, fhashs
from .csv import to_csv, from_csv, merge_in

__author__ = "Zzz, emails: 1174534295@qq.com, 1309458652@qq.com"




from .block import Block
from ..._dz.lists import Lists
# from threading import get_ident as gid
class BlockSocket:
    def __init__(self, skt):
        self.skt = skt
        self.lists = Lists('send', 'recv')
        for k in "add,adds,enable,disable".split(","):
            setattr(self, k, getattr(self.lists, k))
        self.blk = Block()
    def send(self, bts):
        bts = self.lists.send(bts)
        dt=self.blk.wrap(bts)
        #print(f"[{gid()}][BlockSocket] before lists.send")
        #print(f"[{gid()}][BlockSocket] send: {len(dt)}")
        self.skt.send(dt)
    def recv(self, size=1024):
        bts, ch = self.blk.get()
        while len(bts)==0:
            #print(f"[{gid()}][BlockSocket] loop init") 
            bts = self.skt.recv(size)
            #print(f"[{gid()}][BlockSocket] loop recv: {len(bts)}") 
            bts, ch = self.blk.get(bts)
            #print(f"[{gid()}][BlockSocket] loop get: {len(bts), ch}") 
            if len(bts)>0 or ch==0:
                break
        #print(f"[{gid()}][BlockSocket] recv: {len(bts)}")
        bts = self.lists.recv(bts)
        #print(f"[{gid()}][BlockSocket] after lists.recv: {len(bts)}")
        return bts




import sys
from . import fhs
args = sys.argv[1:]
fp = args.pop(0)
hash_type = "sha256"
if len(args)>0:
    hash_type = args.pop(0)
blk_sz = 1024
if len(args)>0:
    blk_sz = int(args.pop(0))
rst = fhs.fhash(fp, hash_type, blk_sz)
print(f"hash value of '{fp}' by '{hash_type}':\n{rst}")



class ObjConf:
    def __str__(self):
        return str(self._maps)
    def __repr__(self):
        return self.__str__()
    def __init__(self):
        self._maps = {}
    def __getitem__(self, key):
        return self._maps[key]
    def __getattr__(self, key):
        if key in self._maps:
            return self._maps[key][0]
        return self.__getattribute__(key)
    def __setattr__(self, key, val):
        if key != '_maps':
            self(key, val)
        else:
            object.__setattr__(self,key,val)
    def __call__(self, *args, **maps):
        '''
            call(key,val)
            or
            call(key0=val0, key1=val1)
        '''
        if len(args)==2:
            key,val = args
            self._maps[key] = val
        else:
            for key,val in maps.items():
                self._maps[key] = val
    def __delattr__(self, key):
        if key in self._maps:
            del self._maps[key]
        return super().__delattr__(key)
    def __contains__(self, key):
        return self._maps.__contains__(key)

pass

class ObjTypeConf(ObjConf):
    def __init__(self, vtype=0):
        super().__init__()
        self._vtypes = {}
        self._vtype = vtype
    def __getitem__(self, vtype=None):
        if vtype is None:
            return self._maps
        maps = {}
        for key in self._maps:
            _vtype = self._vtypes[key]
            if _vtype==vtype:
                maps[key]=self._maps[key]
        return maps
    def __setattr__(self, key, val):
        if key not in {'_maps','_vtype','_vtypes'}:
            self(key, val)
        else:
            object.__setattr__(self,key,val)
    def __call__(self, *args, **maps):
        '''
            call(key,val)
            call(key,val,vtype)
            or
            call(key0=val0, key1=val1)
        '''
        if len(maps)==0:
            if len(args)<=1:
                return self.__getitem__(*args)
        if len(args) in {2,3}:
            key = args[0]
            val = args[1]
            vtype = None
            if len(args)==3:
                vtype = args[2]
            vtype = vtype or self._vtype
            self._maps[key] = val
            self._vtypes[key] = vtype
        else:
            for key,val in maps.items():
                self._maps[key] = val
                self._vtypes[key] = self._vtype
    def __delattr__(self, key):
        if key in self._maps:
            del self._maps[key]
            del self._vtypes[key]
            return
        return super().__delattr__(key)

pass



#

import re
from .jsc import fetch_methods_text, rps_dct, fetch_vals, fetch_vars
from .util import make_fcs
from buildz import argx, xf, dz
'''
public type abc(key, ...){
    ...
}
add:
public type fcs(String keys, Object[] vals, ...) {

}
'''
sc_vals = '''
if (i<vals.length) {
                <get_val><call>(ks[i], vals[i]<others>);
            } else {
                <get_val><call>(ks[i]<null_default><others>);
            }
'''.strip()
sc_nval = '''
<get_val><call>(ks[i]<others>);
'''.strip()
sc_fcs = """
    <type> <method>(<params>)<exp> {
        String[] ks = sptsKeys(keys);
        <align><def_rst>
        for (int i=0;i<ks.length;i++) {
            <def_val>
            <call>
            <add_val>
        }
        <ret_rst>
    }
""".rstrip()
'''
methods(method_name, val, ret, align, null_default)
'''
def _fcs(_type, method, params, exp, args):
    rst = []
    args = xf.loadx(args, out_args=True, as_args=False)
    lst,maps = args.lists, args.dicts
    fetch = argx.Fetch(*xf.loads(r"[method,val, ret,align,default],{}"))
    conf = fetch.fetch(lst, maps)
    wmethod, has_val,ret, align, default = dz.g(conf, method=0, val=None, ret=None,align=True,default='null')

    params = [k.strip() for k in fetch_vars(params)]
    if len(params)==1:
        has_val = False
    if has_val:
        param = params[1]
        val_type = ' '.join(param.split(" ")[:-1])
    else:
        val_type = None
        align = False
    vars = [k.split(" ")[-1].strip() for k in params]
    if len(params)==1:
        val_type = None
        align = False
    unit_type = _type.replace("public","").replace("private","").replace("protected","").replace("static","").strip()
    if ret is None:
        ret = unit_type not in ('void','')
    check_align = ''
    if align:
        assert val_type is not None, f"_fcs({_type}, {method}, {params}, {args})"
        check_align = 'if(ks.length>vals.length)throw new RuntimeException("not enough vals");'
    if (val_type is not None):
        outs_params = ["String keys", f"{val_type}[] vals"]+params[2:]
        others = vars[2:]
        sc_v = sc_vals
    else:
        sc_v = sc_nval
        outs_params = ["String keys"]+params[1:]
        others = vars[1:]
    if len(others)==0:
        others = ""
    else:
        others = ", ".join(others)
        others = ", "+others
    outs_params = ", ".join(outs_params)
    null_default = ""
    if default is not None:
        null_default = f", {default}"
    out_type = "public void"
    def_rst, ret_rst, def_val, get_val, add_val = "","","","",""
    if ret:
        out_type = f"public {unit_type}[]" 
        def_rst = f"{unit_type}[] rst = new {unit_type}[ks.length];" 
        ret_rst = "return rst;"
        def_val = f"{unit_type} val;"
        get_val = "val="
        add_val = "rst[i] = val;"
    sc_v = rps_dct(sc_v, get_val=get_val, call=method, others = others, null_default = null_default)
    s = rps_dct(sc_fcs, type=out_type, method = wmethod, params = outs_params, align = check_align, def_rst = def_rst, def_val = def_val, call = sc_v, add_val = add_val, ret_rst = ret_rst, exp=exp)
    return [s], []
fcs = make_fcs(_fcs, False)



import struct
class Block:
    '''
        在网络流数据上使用格式"数据长度+数据"，数据长度用4字节整数
    '''
    def __init__(self):
        self.caches = b""
        self.build = self.wrap
    def empty(self):
        return len(self.caches)==0
    def clean(self):
        self.caches = b''
    def get(self, bts=b""):
        '''
            返回数据和缓存剩余大小
            缓存数据不足以返回单个数据时，返回的数据为b''
        '''
        if bts!=b'':
            self.caches+=bts
        lc = len(self.caches)
        if lc<4:
            return b"", lc
        bsz = self.caches[:4]
        sz = struct.unpack("<I", bsz)[0]
        if lc<sz+4:
            return b'', lc
        dts = self.caches[4:4+sz]
        self.caches = self.caches[4+sz:]
        return dts, len(self.caches)
    def wrap(self, bts):
        '''
            输入数据
            返回"数据长度+数据"，数据长度4字节
        '''
        sz = len(bts)
        bsz = struct.pack("<I", sz)
        return bsz+bts

pass
